#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/hub/test_passfd.py $
$Id: test_passfd.py 26882 2005-06-02 14:41:05Z dbinger $
"""

import os, sys, socket
from qp.hub import passfd

#
# Create a pipe for sending the fd.
#

if __name__ == "__main__":
    rfd, wfd = passfd.socketpair(socket.AF_UNIX, socket.SOCK_STREAM)
    print "rfd", rfd, "wfd", wfd

    pid = os.fork()

    if pid != 0:
        # We're in the parent.

        # Open a file for passing along to child.

        fileObj = open('/dev/zero', 'r')

        # ioctl() will only pass raw filedescriptors. Find fd of fileObj.
        fd = fileObj.fileno()

        # Send to the child
        os.write(wfd, "x")
        passfd.sendfd(wfd, fd)

        # Wait for child to terminate, then exit.
        os.waitpid(pid, 0)
        sys.exit(0)

    else:
        # We're in the child.

        print os.read(rfd, 1)
        fd = passfd.recvfd(rfd)

        # Reopen the filedescriptor as a Python File-object.
        fileObj = os.fdopen(fd, 'r')

        # Example usage: Read file, print the first line.
        print "first byte", `fileObj.read(1)`
        sys.exit(0)




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/profiler.py $
$Id: profiler.py 28413 2006-05-25 19:36:51Z dbinger $
"""
from cStringIO import StringIO
from profile import runctx
from pstats import Stats
from qp.pub.common import get_publisher
from tempfile import NamedTemporaryFile
import sys


class Profile (object):

    def __init__(self, PATH_INFO='/', user=None, content='', **env):
        publisher = get_publisher()
        if isinstance(user, basestring):
            user = publisher.get_users().get(user)
        if user:
            session_id = 'bogus'
            site = publisher.get_site().get_name()
            env.setdefault('HTTP_COOKIE', '%s=%s' % (site, session_id))
            def abort():
                publisher.__class__.abort(publisher)
                session = publisher.create_session()
                session.owner = user
                session.effective_user = user
                get_publisher().get_sessions()[session_id] = session
            publisher.abort = abort
        env['PATH_INFO'] = PATH_INFO
        env.setdefault('SCRIPT_NAME', '')
        env.setdefault('HTTPS', 'on')
        env.setdefault('SERVER_NAME', 'example.org')
        if content:
            env.setdefault('REQUEST_METHOD', 'POST')
        else:
            env.setdefault('REQUEST_METHOD', 'GET')
        def commit():
            def ignore(*args, **kwargs):
                pass
            publisher.connection.storage.end = ignore
            publisher.__class__.commit(publisher)
            publisher.__class__.abort(publisher)
        publisher.commit = commit
        input = StringIO(content)
        output = StringIO()
        def process():
            hit = publisher.process(input, env)
            hit.get_response().write(output)
            return hit, output.getvalue()
        tmpfile = NamedTemporaryFile(suffix=".profile")
        context = dict()
        runctx("hit, page = process()", locals(), context, tmpfile.name)
        self.stats = Stats(tmpfile.name)
        self.stats.sort_stats('time', 'calls')
        tmpfile.close()
        self.hit = context.get('hit')
        self.response_string = context.get('page')
        if 'abort' in vars(publisher):
            del publisher.abort
        del publisher.commit

    def get_stats(self):
        return self.stats

    def get_response_string(self):
        return self.response_string

    def get_hit(self):
        return self.hit

    def get_status_code(self):
        return self.get_hit().get_response().get_status_code()

    def sort(self, *columns):
        self.stats.sort_stats(*columns)

    def format(self, *restrictions):
        stdout = sys.stdout
        output = StringIO()
        try:
            sys.stdout = output
            self.stats.print_stats(*restrictions)
        finally:
            sys.stdout = stdout
        return output.getvalue()

    def calls(self, function_name):
        return [value[0]
            for key, value in self.stats.stats.items()
            if key[-1] == function_name]

    def loads(self):
        return sum(self.calls('_p_load_state'))

    def __str__(self):
        self.sort('time', 'calls')
        return self.format(10)

    def callers(self, *args):
        self.stats.print_callers(*args)

    def callees(self, *args):
        self.stats.print_callees(*args)




#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/test/utest_util.py $
$Id: utest_util.py 26453 2005-04-01 02:15:29Z dbinger $
"""
from qp.lib.util import import_object
from sancho.utest import UTest

class Test(UTest):

    def check_import_object(self):
        import sys
        assert sys is import_object('sys')
        assert UTest is import_object('sancho.utest.UTest')

if __name__ == '__main__':
    Test()



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/test/utest_delegation.py $
$Id: utest_delegation.py 28598 2006-08-16 19:16:51Z dbinger $
"""
from qp.lib.delegation import delegate, delegate_passing_self
from sancho.utest import UTest

class Test(UTest):

    def a(self):
        class A(object):
            def get_args1(self, *args, **kwargs):
                return (self, args, kwargs)
            def get_args2(self, *args, **kwargs):
                return (self, args, kwargs)
        class B(object):
            def __init__(self):
                self.c = A()
                self.d = A()
        delegate(B, 'c.get_args1')
        delegate_passing_self(B, 'd.get_args2')
        b = B()
        assert b.get_args1() == (b.c, (), {}), b.get_args1()
        assert b.get_args2() == (b.d, (b,), {})

if __name__ == '__main__':
    Test()




BASIC_FORM_CSS = """\
div.qpform {
    float: left;
}

div.qpform div.widget,
div.qpform div.submit {
    clear: left;
    padding: 0.5ex;
    margin: 0.5ex;
}

div.qpform div.widget div.widget,
div.qpform div.submit div.widget {
    margin: 0;
    padding: 0;
    clear: none;
}

div.qpform div.title {
    font-weight: bold;
}

div.qpform div.close {
    clear: left;
}

div.qpform div.submit div.widget {
    clear: none;
    float: left;
    margin-right: 0.5ex;
}

div.qpform div.error {
    color: #c30;
    font-size: small;
    margin-top: 0.2ex;
}

div.qpform div.hint {
    font-size: small;
    font-style: italic;
    margin-top: 1ex;
    width: 40ex;
}

div.qpform div.errornotice {
    color: #c30;
    padding: 1ex;
    margin: 1ex;
    width: 40ex;
}
"""



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/publish.py $
$Id: publish.py 28833 2006-09-21 16:07:44Z dbinger $
"""
from datetime import datetime
from durus.btree import BTree
from durus.connection import Connection
from durus.error import ConflictError
from durus.file_storage import TempFileStorage, FileStorage
from durus.storage_server import StorageServer
from os import getpid
from pprint import pformat
from qp.fill.directory import Directory
from qp.fill.form import Form
from qp.hub.web import run_web
from qp.lib.site import Site
from qp.lib.spec import spec, add_getters, specify, match
from qp.lib.tz import UTC
from qp.lib.util import randbytes
from qp.mail.send import Email
from qp.pub.common import get_hit, set_publisher, get_request, get_response
from qp.pub.common import get_user, get_session
from qp.pub.hit import Hit
from qp.pub.session import Session
from qp.pub.user import User, compute_digest
from qpy import h8
from socket import getfqdn
from sys import exc_info
from time import strptime, mktime
from traceback import format_exception
from urlparse import urljoin
import binascii
import qp.lib.site

class RespondNow (Exception):
    """
    This Exception is used to break out of the path traversal.
    The Publisher expects the contents of the response to be
    set before this exception is raised.
    """

class Publisher (object):
    """
    The basic QP Publisher class.  This is implements the normal traversal,
    page generation, and error handling pattern used in QP.  Subclasses may
    override fill_response() or other methods to customize this as desired.
    """
    site_is = spec(
        Site,
        "the QP Site")
    root_directory_is = spec(
        Directory,
        "The root directory")
    hit_is = spec(
        Hit,
        "The Hit (Request, Response, Session, etc.) currently being "
        "processed, or None.")

    test_site_name = "echo"

    def __init__(self, site=None, **other):
        set_publisher(self)
        self.set_hit(None)
        specify(self, site=site or Site(self.test_site_name))
        self.root_directory = None

    def run_web(site):
        """
        Launch the web server(s) for this site.
        The default implementation uses the qp web server to
        provide http and scgi service if addresses for those
        services are present in self.configuration.
        """
        run_web(site)
    run_web = staticmethod(run_web)

    def __call__(self, wsgi_env, wsgi_start_response):
        """(wsgi_env:dict, wsgi_start_response:callable)
        This makes it possible to use a Publisher instance as a
        wsgi application.
        """
        input = wsgi_env['wsgi.input']
        hit = self.process(input, wsgi_env)
        response = hit.get_response()
        status = "%03d %s" % response.get_status()
        headers = response.generate_headers()
        wsgi_start_response(status, headers)
        return response.generate_body_chunks()

    def process(self, stdin, env):
        """
        Return a Hit with a complete response.
        Subclasses that want to customize the Hit class may do so by overriding
        this method.
        """
        if env.get('SCRIPT_NAME') and env.get('PATH_INFO') == '':
            # The "script" here is your QP application, so the SCRIPT_NAME
            # should be the part of the url, after the host name, that
            # causes requests to be handled by your QP application.
            # The PATH_INFO part of the path after the SCRIPT_NAME.
            # We expect applications running at the top of the URL space
            # to get requests where the SCRIPT_NAME is the empty string
            # and the PATH_INFO starts with a '/'.

            # Your HTTP server is not providing the expected CGI environment.
            # The full path part of the URL is being put in the SCRIPT_NAME.
            # We'll try to make it work anyway, but we are assuming here
            # that your application is running at the top of your URL space.
            # If you want your application to run somewhere else in your URL
            # space, you must change your server so that it provides the
            # correct environment variables.  Alternatively, your publisher
            # class can override this method and hack the environment as
            # needed to fit QP's expectations.
            env['PATH_INFO'] = env['SCRIPT_NAME']
            env['SCRIPT_NAME'] = ""
        hit = Hit(stdin, env)
        self.process_hit(hit)
        return hit

    def process_hit(self, hit):
        """(hit:Hit)
        This processes the request on the hit and sets the response.
        Subclasses that want to customize error handling may do so by
        overriding this method.
        """
        self.set_hit(hit)
        hit.get_request().process_inputs()
        try:
            hit.init_response()
            self.fill_response()
        except RespondNow:
            self.log_hit()
        except:
            self.handle_exception()
        self.log_hit()

    def fill_response(self):
        """
        Set the response.  Subclasses that want to use some other
        template system may do so by overriding this method.
        """
        self.fill_response_using_root_directory()

    def set_hit(self, hit):
        """(hit:Hit)
        Set the current Hit.
        """
        self.hit = hit

    def respond_now(self):
        """
        This breaks out of path traversal.  When this is called, the
        response should have already been prepared for sending.
        """
        raise RespondNow

    def handle_exception(self):
        """
        This is called when there is an error exception while processing
        a request.
        """
        self.log_exception()
        if exc_info()[0] is SystemExit:
            raise
        self.get_hit().init_response()
        self.get_hit().get_response().set_status(500)
        if self.display_exceptions():
            self.display_exception()
        else:
            self.hide_exception()

    def secure(self):
        """If the scheme is not https, redirect so that it will be.
        """
        if (get_request().get_scheme() != 'https' and
            self.get_site().get_https_address()):
            self.redirect(self.complete_url('', secure=True))

    def complete_url(self, path, secure=False):
        """(path:str, secure:bool=False) -> str
        Turn path into a complete url to this publisher, changing the
        scheme to https if secure is True.
        """
        s = str(path)
        if not secure and s.startswith('http://'):
            return s
        if s.startswith('https://'):
            return s
        if s.startswith('/'):
            complete_path = get_request().get_script_name() + s
        else:
            complete_path = s
        if secure:
            host, port = self.get_site().get_https_address()
            if not host:
                host = get_request().get_server().split(':')[0]
            if port == 443:
                address = str(host)
            else:
                address = "%s:%s" % (host, port)
            base = 'https://%s%s' % (address,
                                     get_request().get_path_query())
        else:
            base = get_request().get_url()
        return urljoin(base, complete_path)

    def redirect(self, location, permanent=False):
        """(location:str, permanent:boolean=False)
        This prepares a redirect response and uses an exception
        to break out of the path traversal and return the response
        immediately.
        """
        if permanent:
            status = 301
        else:
            status = 302
        response = self.get_hit().get_response()
        response.set_status(status)
        response.set_header('location', self.complete_url(location))
        response.set_content_type('text/plain', 'iso-8859-1')
        response.set_body(
            "Your browser should have redirected you to %s" % location)
        self.respond_now()

    def not_found(self, body=None):
        """(body:str)
        This prepares a 404 response and uses an exception
        to break out of the path traversal and return the response
        immediately.
        """
        self.respond('Not Found', body or 'That page is not here.', status=404)

    def is_live_host(self):
        return self.get_site().get_live_host_name() == getfqdn()

    def log_exception(self):
        """
        This is called to record an error exception.
        """
        report = self.format_exception_report()
        print report
        if self.is_live_host():
            self.send_to_administrator(report)

    def display_exception(self):
        """
        This places an exception report in the response.
        """
        get_hit().get_response().set_body(self.format_exception_report())
        get_hit().get_response().set_content_type('text/plain', 'iso-8859-1')

    def format_user_info_for_log(self):
        return '- -'

    def format_log_hit_line(self):
        """() -> str
        """
        hit = get_hit()
        code = hit.get_response().get_status_code()
        duration = str(datetime.utcnow() - hit.get_time()).lstrip('0:')
        request = hit.get_request()
        path_query = request.get_path_query()
        method = request.get_method()
        remote = request.get_remote_address()
        scheme = request.get_scheme()
        try:
            user = self.format_user_info_for_log()
        except:
            user = "! !"
            print "[!%s in format_user_info_for_log()]" % exc_info()[0]
        referrer = request.get_header('http-referer', '-').replace(' ', '_')
        agent = request.get_header('user-agent', '-').replace(' ', '_')
        time = hit.get_time().strftime("%Y-%m-%d %H:%M:%S")
        pid = getpid()
        msg = ('%(time)s %(code)s %(duration)s %(remote)s %(user)s %(pid)s '
               '%(scheme)s %(method)s %(path_query)s %(referrer)s %(agent)s' %
               locals())
        return msg

    def parse_log_hit_line(self, line):
        """(str) -> dict | None
        If the line matches the format produced by format_log_hit_line(),
        this returns a dictionary of component values.  This is intended
        to make it easier to write scripts to produce access logs in
        alternative formats.
        """
        parts = line.split(' ')
        if len(parts) != 13:
            return None
        try:
            time_tuple = strptime(line[:19], "%Y-%m-%d %H:%M:%S")
        except:
            return None
        return dict(time=datetime.fromtimestamp(mktime(time_tuple)),
                    status_code=parts[2],
                    duration=float(parts[3]),
                    remote_ip=parts[4],
                    owner=parts[5],
                    user=parts[6],
                    pid=parts[7],
                    scheme=parts[8],
                    method=parts[9],
                    path=parts[10],
                    referrer=parts[11],
                    agent=parts[12].replace('_', ' '))

    def log_hit(self):
        """
        This logs the processing of a request.
        """
        print self.format_log_hit_line()

    def format_exception_report(self):
        """() -> str
        This returns a string that reports an error exception.
        """
        hit = get_hit()
        request = hit.get_request()
        report = ''.join(format_exception(*exc_info()))
        report += '\npath = %r' % request.get_path()
        report += '\nquery= %r' % request.get_query()
        report += '\n\ninfo:\n' + pformat(hit.get_info())
        report += '\n\nvars(request):\n' + pformat(vars(request))
        report += '\n'
        return report

    def hide_exception(self):
        """
        This sets the response to be a page that can be shown when there
        is an error exception, but you don't want to expose the details
        of the exception.
        """
        get_hit().get_response().set_status(500)
        get_hit().get_response().set_body(
            self.page('Regrets',
                      h8("<p>This page is temporarily unavailable.</p>"
                         "<p>Please try again later.</p>")))

    def get_root_directory(self):
        if self.root_directory is None:
            self.root_directory = self.site.get_root_directory_class()()
        return self.root_directory

    def fill_response_using_root_directory(self):
        """
        Traverse the components, set the response body, using the
        normal _q_traverse() method.
        """
        path = get_request().get_path_info()
        components = path[1:].split('/')
        body = self.get_root_directory()._q_traverse(components)
        get_response().set_body(body)

    def display_exceptions(self):
        """() -> bool
        Should the details of error exceptions be reported in responses?
        """
        return not self.is_live_host()

    def respond(self, title, *content, **kwargs):
        """(title:str, *content, **kwargs)
        Fill the response using the page() method and the given title,
        content, and **kwargs and return immediately.
        """
        self.get_hit().init_response()
        if 'status' in kwargs:
            status = kwargs['status']
            del kwargs['status']
        else:
            status = 400
        self.get_hit().get_response().set_status(status)
        self.get_hit().get_response().set_body(
            self.page(title, *content, **kwargs))
        self.respond_now()

    # Email-related methods.

    def get_webmaster_address(self):
        """() -> email_address:str
        """
        return 'webmaster@%s' % getfqdn()

    def get_debug_address(self):
        """() -> email_address:str|None

        If this returns a nonempty string, the qp.mail.send.Email
        class will use this as the smtp recipient for all messages it
        sends.  If you want email sent by that class to go to the normal
        recipients, you must override this to return None.
        """
        if self.is_live_host():
            return None
        else:
            return self.get_webmaster_address()

    def is_email_enabled(self):
        """() -> bool
        The qp.mail.send.Email class will never send any email unless this
        returns true.  If you want any email sent (using that class),
        you must override this to return True.
        """
        return False

    def get_smtp_server(self):
        """() -> str
        """
        return 'localhost'

    def send_to_administrator(self, message):
        """(message:str)
        """
        if self.is_email_enabled():
            email = Email()
            email.set_subject(self.get_site().get_name())
            email.set_body(message)
            if email.send():
                print '\nemail:sent'

    # Page formatting methods.

    def header(self, title, doctype='xhtml1-strict', style=None, **kwargs):
        """(title, **kwargs) -> h8
        Return the site-standard html header.
        """
        if doctype == 'xhtml1-transitional':
            s = h8(
                '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"'
                ' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">')
        elif doctype == 'xhtml1-strict':
            s = h8(
                '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"'
                ' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">')
        else:
            s = h8()
        s += h8('<html xmlns="http://www.w3.org/1999/xhtml" '
                'lang="en" xml:lang="en">')
        s += h8('<head>')
        if style is not None:
            s += h8('<style type="text/css">%s</style>') % style
        s += h8('<title>%s</title>')  % title
        s += h8('</head><body>')
        return s

    def footer(self, *args, **kwargs):
        """(**kwargs) -> h8
        Return the site-standard html footer.
        """
        return h8('</body></html>')

    def page(self, title, *content, **kwargs):
        """(title, *content, **kwargs) -> h8
        Return a page formatted according to the site-standard.
        """
        return (self.header(title, **kwargs) +
                h8.from_list(content) +
                self.footer(title=title, **kwargs))

    def get_time_zone(self):
        return UTC

add_getters(Publisher)


class DurusPublisher (Publisher):
    """
    This is the standard Publisher class for QP sites that use
    a Durus Database.
    """

    connection_is = spec(
        Connection,
        "the Durus Connection")

    test_site_name = "proto"

    def __init__(self, connection=None, **other):
        Publisher.__init__(self, **other)
        self.connection = connection or Connection(TempFileStorage())
        self.ensure_database_initialized()

    def ensure_database_initialized(self):
        """
        Make sure that database root has sessions and users mappings.
        If not, make them and commit().
        """
        self.ensure_sessions_initialized()
        self.ensure_users_initialized()

    def ensure_sessions_initialized(self):
        if self.get_sessions() is None:
            self.get_root()['sessions'] = BTree()
            self.commit()

    def ensure_users_initialized(self):
        if self.get_users() is None:
            users = BTree()
            users[''] = self.create_user("")
            self.get_root()['users'] = users
            self.commit()

    def run_durus(site):
        """
        Launch the durus server for this site.
        """
        StorageServer(FileStorage(site.get_durus_file()),
            address=site.get_durus_address()).serve()
    run_durus = staticmethod(run_durus)

    def create_user(self, user_id):
        return User(user_id)

    def create_session(self):
        return Session()

    def commit(self):
        """Commit changes to the database."""
        self.get_connection().commit()

    def abort(self):
        """
        Make sure that the connection has no uncommitted object state.
        """
        self.get_connection().abort()

    def get_root(self):
        """() -> PersistentDict
        Return the root object of the database.
        """
        return self.get_connection().get_root()

    def get_sessions(self):
        """() -> BTree
        Return the session mapping.
        """
        return self.get_root().get('sessions')

    def get_users(self):
        """() -> BTree
        Returns the mapping from user ids to user instances.
        """
        return self.get_root().get('users')

    def add_user(self, user):
        """(User)
        Add a new user to the application.
        """
        users = self.get_users()
        assert user.get_id() not in users
        users[user.get_id()] = user

    def lookup_user(self, id=None, email=None, **kwargs):
        """
        Returns a User or None if no such user is found.
        """
        user = self.get_users().get(id)
        if user is not None:
            return user
        if (email is not None and
            match(email, self.get_users().get('').email_is)):
            lower_email = email.lower()
            for user in self.get_users().itervalues():
                user_email = user.get_email()
                if user_email is not None and user_email.lower() == email:
                    return user
        return None

    # Authentication methods

    def sign_out(self, url):
        """(url:str)
        Un-authenticate the current user and redirect to `url`.
        """
        get_session().clear_authentication()
        self.redirect(url)

    def ensure_signed_in_using_form(self, title='Please Sign In',
                                    realm=None, **kwargs):
        """
        Make sure that the current user is signed in.
        This presents a form to the user.  Because the form transmits the
        password, this redirects to the https address before presenting the
        form.

        The realm, if given, identifies which of the user's passwords
        should be used.  The default realm is the name of the site.
        """
        if not get_user():
            self.secure()
            form = Form(use_tokens=False)
            default_user_id = None
            if get_session().get_owner():
                default_user_id = get_session().get_owner().get_id()
            form.add_string(
                'name', value=default_user_id, title="User Name", required=True)
            form.add_password('password', title="Password", required=True)
            form.add_submit('login', 'Sign in')
            def show_form():
                self.respond(title, form.render(), status=403, **kwargs)
            if not form.is_submitted()or form.has_errors():
                show_form()
            user = self.get_users().get(form.get('name'))
            if user and user.has_password(form.get('password') or '',
                                          realm=realm):
                get_session().set_authenticated(user)
                self.redirect('')
            if not user:
                form.set_error('name', 'unknown user')
            else:
                form.set_error('password', 'wrong password')
            show_form()

    def ensure_signed_in_using_basic(self, realm=None):
        """
        Make sure that the current user is signed in.
        This uses HTTP Basic Authentication.  Because Basic Authentication
        transmits the password, this implementation redirects if necessary
        to make sure that the challenge only happens when the scheme is
        https.

        The realm, if given, identifies which of the user's passwords
        should be used.  The default realm is the name of the site.
        """
        if not get_user():
            # Look to see if authentication credentials have been delivered.
            authorization = get_request().get_header(
                'HTTP_AUTHORIZATION', '').split()
            if len(authorization) == 2:
                scheme, encoded = authorization
                try:
                    decoded = binascii.a2b_base64(encoded).split(':')
                except binascii.Error:
                    pass
                else:
                    if scheme.lower() == 'basic' and len(decoded) == 2:
                        username, password = decoded
                        user = self.get_users().get(username)
                        if user and user.has_password(password, realm=realm):
                            get_session().set_authenticated(user) # success
        if not get_user():
            # Issue an authentication challenge.
            assert self.get_site().get_https_address(), (
                "If you want basic authentication, use https.")
            self.secure() # This redirects to https.
            if realm is None:
                realm = self.get_site().get_name()
            self.get_hit().init_response()
            self.get_hit().get_response().set_status(401)
            self.get_hit().get_response().set_header(
                'WWW-Authenticate',
                'Basic realm="%s"' % realm)
            self.respond_now()

    def ensure_signed_in_using_digest(self, realm=None):
        """
        Make sure that the current user is signed in.
        This uses HTTP Digest Authentication.  Since this does not transmit
        the password itself, we allow this to work even when https is not
        available.  If https is available, though, we redirect to it first.

        The realm, if given, identifies which of the user's passwords
        should be used.  The default realm is the name of the site.
        """
        if not get_user():
            if realm is None:
                realm = self.get_site().get_name()
            def attempt_digest_authentication():
                authorization = get_request().get_header(
                    'HTTP_AUTHORIZATION', '').split()
                if len(authorization) < 5:
                    return
                if authorization[0].lower() != 'digest':
                    return
                parameters = {}
                for item in authorization[1:]:
                    item_split = item.split('=')
                    if len(item_split) != 2:
                        continue
                    name, value = item_split
                    if value[-1] == ',':
                        value = value[:-1]
                    if value and value[0]=='"' and value[-1]=='"':
                        value = value[1:-1]
                    parameters[name.lower()] = value
                username = parameters.get('username', None)
                if not username:
                    return
                user = self.get_users().get(username)
                if not user:
                    return
                nonce = parameters.get('nonce')
                if nonce not in get_user().get_tokens():
                    return
                get_user().get_tokens().remove(nonce)
                cnonce = parameters.get('cnonce')
                if not cnonce:
                    return
                nc = parameters.get('nc')
                if not nc:
                    return
                qop = parameters.get('qop')
                if not qop:
                    return
                request_digest = parameters.get('response', '')
                if len(request_digest) != 32:
                    return
                realm = parameters.get('realm')
                user_digest = user.get_digester().get_digest(realm)
                if not user_digest:
                    return
                method = get_request().get_method()
                if request_digest == compute_digest(
                    user_digest,
                    nonce,
                    nc,
                    cnonce,
                    qop,
                    compute_digest(method, parameters.get('uri', ''))):
                    # success
                    get_session().set_authenticated(user)
            attempt_digest_authentication()
        if not get_user():
            # Issue an authentication challenge.
            self.secure()
            self.get_hit().init_response()
            self.get_hit().get_response().set_status(401)
            self.get_hit().get_response().set_body("Do I know you?")
            nonce = get_user().get_tokens().new_token()
            self.get_hit().get_response().set_header(
                'WWW-Authenticate',
                ('Digest realm="%s", nonce="%s", opaque="0%s", stale=false, '
                 'algorithm=MD5, qop="auth"' % (realm, nonce, nonce)))
            self.respond_now()

    def ensure_signed_in(self, **kwargs):
        """
        Make sure the user is authenticated using the default method.
        """
        self.ensure_signed_in_using_digest()

    def process_hit(self, hit):
        """(hit:Hit)
        This processes the request on the hit and sets the response
        on the hit.  It handles conflict errors and commits changes
        when there are no error exceptions.  Note that this overrides
        a method of the base class.
        """
        self.set_hit(hit)
        try:
            hit.get_request().process_inputs()
            for attempt in range(5):
                try:
                    hit.init_response()
                    self.abort()
                    try:
                        self.fill_response()
                    except RespondNow:
                        pass
                    self.log_hit()
                    self.commit()
                    return
                except ConflictError:
                    self.log_exception()
                    self.abort()
            else:
                raise RuntimeError("too many Conflict errors")
        except:
            self.handle_exception()
            self.log_hit()
            self.abort()

    def fill_response(self):
        """
        Use the request to prepare the response.
        This method takes care of session management, and calls the
        fill_response_with_session_present() method to do the rest.
        Note that this overrides a method of the base class.
        """
        cookie_name = self.get_site().get_name()
        cookie = get_request().get_cookie(cookie_name)
        session = self.get_sessions().get(cookie)
        if session and not session.is_valid():
            del self.get_sessions()[cookie]
            session = None
        if session is None:
            session = self.create_session()
        self.get_hit().set_session(session)
        session.access()
        try:
            self.fill_response_using_root_directory()
        finally:
            session = self.get_hit().get_session()
            if session.needs_saving():
                cookie = randbytes(8)
                self.get_sessions()[cookie] = session
                get_response().set_cookie(
                    cookie_name, cookie,
                    path=get_request().get_script_name() or "/",
                    secure=(get_request().get_scheme() == 'https'))

    def format_user_info_for_log(self):
        def format_id(user):
            user_id = user.get_id()
            if  user_id == '':
                return '-'
            try:
                return str(user_id).replace(' ', '_')
            except UnicodeDecodeError:
                return repr(user_id).replace(' ', '_')
        session = get_session()
        owner = session.get_owner()
        owner_part = format_id(owner)
        user = session.get_effective_user()
        if user.get_id() != '' and user is owner:
            user_part = '+'
        else:
            user_part = format_id(user)
        return owner_part + " " + user_part

add_getters(DurusPublisher)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/test/utest_user.py $
$Id: utest_user.py 28032 2006-03-07 17:07:14Z dbinger $
"""
from qp.pub.common import clear_publisher
from qp.pub.publish import Publisher
from qp.pub.user import User, Permissions
from sancho.utest import UTest

class Test(UTest):

    def _pre(self):
        clear_publisher()

    def _post(self):
        clear_publisher()

    def check_a(self):
        publisher = Publisher()
        user = User('a')
        assert user
        assert not User('')
        assert user.get_id() == 'a'
        assert user.get_digester().get_realms() == []
        user.set_password('foo')
        assert user.get_digester().get_realms()
        assert user.has_password('foo')
        assert not user.has_password('bah')
        user.set_password('')
        assert user.get_digester().get_realms() == []
        assert isinstance(user.get_permissions(), Permissions)
        assert not user.is_granted('power')
        assert not user.is_admin()
        try:
            user.get_permissions().grant('administrator', None)
            assert 0
        except TypeError:
            pass
        user.get_permissions().grant('administrator', user)
        assert not user.is_admin()
        user.get_permissions().ungrant('administrator', user)
        user.get_permissions().grant('administrator', user)
        user.get_permissions().grant('administrator', True)
        assert user.is_admin()
        user.get_permissions().ungrant('administrator', True)
        assert not user.is_admin()
        user.get_permissions().grant('power', user)
        assert not user.is_granted('power')
        assert user.is_granted('power', user)
        token = user.get_tokens().new_token()
        assert token in user.get_tokens()
        user.get_tokens().discard(token)
        assert token not in user.get_tokens()
        user.get_tokens().discard(token)
        user.get_tokens().clear()

if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/fill/test/utest_html.py $
$Id: utest_html.py 28351 2006-05-10 17:27:02Z dbinger $
"""
from sancho.utest import UTest
from qp.fill.html import href, div, ol, ul, dl, nl2br, url_with_query
from qpy import h8

class HTMLTest(UTest):

    def a(self):
        result = href('a', 'b', name="c")
        assert isinstance(result, h8)
        assert result == '<a href="a" name="c">b</a>'

    def b(self):
        result = div('a', 'b', c="d")
        assert isinstance(result, h8)
        assert result == '<div c="d">ab</div>'

    def c(self):
        result = ol(['a', 'b'], c="d")
        assert isinstance(result, h8)
        assert result == '<ol c="d"><li>a</li><li>b</li></ol>'

    def d(self):
        result = ul(['a', 'b'], c="d")
        assert isinstance(result, h8)
        assert result == '<ul c="d"><li>a</li><li>b</li></ul>'

    def e(self):
        result = dl([('a', 'b'), ('e', None)], c="d")
        assert isinstance(result, h8)
        assert result == '<dl c="d"><dt>a</dt><dd>b</dd><dt>e</dt></dl>'

    def f(self):
        result = nl2br('')
        assert isinstance(result, h8)
        result = nl2br('a\nb\n')
        assert isinstance(result, h8)
        assert result == 'a<br />b<br />', result

    def test_url_with_query(self):
        result = url_with_query('/foo', x=1, y='A<')
        assert isinstance(result, h8)
        assert result == '/foo?y=A%3C&x=1', result

if __name__ == '__main__':
    HTMLTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/fill/directory.py $
$Id: directory.py 28352 2006-05-10 17:27:26Z dbinger $
"""
import sys
from qp.pub.common import get_request, get_path, redirect, not_found, get_path
from qp.pub.common import get_response
from urlparse import urljoin


class Directory (object):
    """
    Subclasses of this class define the publishing traversal through
    one '/' of a url.  Usually, this involves overriding get_exports()
    and providing a method for each name that you want to export.
    Note that '' can be, and usually is, exported like any other.
    If your class includes a _q_index() method to provide the response
    for urls that end at this '/', then the result of get_exports() should
    include ('', '_q_index', None, None).
    """

    def get_exports(self):
        """Returns an iterable of 4 tuples:
        (path_component:str,
         attribute_name:str|None,
         crumb:str|None,
         title:str|None)

        Subclasses normally override this to provide the explicit exports.

        The crumb and title parameters allow you to specify what links to this
        export should look like.  The crumb value is used as the content of the
        link.  The title value is use as the title attribute of the link.

        When the attribute_name is None, the _q_lookup should provide a result
        for that path_component.
        """
        return []

    def _q_translate(self, component):
        """(component : string) -> string|None

        Returns the name of the attribute exported for the given path component,
        or None.  Subclasses will not normally need to override this.
        """
        for export, name, crumb, title in self.get_exports():
            if component == export:
                return name
        return None

    def _q_lookup(self, component):
        """(component : string) -> object

        Lookup a path component and return the corresponding object
        (usually a Directory, a method or a string).
        Returning None signals that the component does not exist.

        Subclasses should override this when they want to export path
        components that, for one reason or another, don't have corresponding
        attributes that can be found using _q_translate.  This happens,
        for example, when you want to export the set of all integers or other
        large sets.
        """
        return None

    def _q_traverse(self, path):
        """(path: [string]) -> None|Stream|object

        Traverse a path and return the result.  This is what the
        Publisher calls.  The result will be used to set the body of
        the http response.  If the result is None, the Publisher will
        use not_found() to set the 404 response.  If the result is an
        instance of Stream, it will be written out in in chunks.
        Otherwise, the str() of the result of this call will be used
        as the body of the http response.

        Subclasses don't normally need to override this, but may do so.
        This may be useful to provide access control or to implement
        directories that process the entire path themselves.
        """
        component = path[0]
        name = self._q_translate(component)
        if name is None:
            obj = self._q_lookup(component)
        else:
            obj = getattr(self, name)
        if obj is None:
            not_found('%r?' % component)
        if path[1:]:
            if hasattr(obj, '_q_traverse'):
                result = obj._q_traverse(path[1:])
            else:
                not_found()
        elif callable(obj):
            result = obj()
        else:
            result = obj
        if result is None:
            not_found()
        else:
            return result

    def __call__(self):
        """() -> None|Stream|object
        If the _q_traverse() of a Directory reaches this instance,
        and there are no more components, it calls this instance.
        We could return None to cause a 404 response, but we'll try
        adding a slash to the end of the url if it looks like that
        appears to be the intent of the current request.
        """
        if (self._q_translate('') is not None and
            not get_request().get_fields()):
            # Fix missing trailing slash.
            return redirect(get_request().get_path_info() + "/", permanent=True)
        else:
            return None


def get_directory_path():
    """() -> [object]
    Return the list of traversed instances.
    Thus uses the frame stack, so it is only meaningful in the context of a call
    to _q_traverse.
    """
    path = []
    frame = sys._getframe()
    while frame:
        if frame.f_code.co_name == '_q_traverse':
            self = frame.f_locals.get('self', None)
            if path[:1] != [self]: # avoid repetition
                path.insert(0, self)
        frame = frame.f_back
    return path

def get_path_directory_list():
    """() -> [(relative_path:str, directory:Directory)]
    Provides a relative path for each directory currently traversed.
    """
    app_root = get_request().get_environ('SCRIPT_NAME', '')
    path = get_path()
    path_components = path[len(app_root):].split('/')
    namespace_stack = get_directory_path()
    if get_response().get_status_code() == 200:
        offset = -1
    else:
        offset = 0
    return [('./' + '../' * (len(namespace_stack) - k + offset),
             namespace_stack[k])
            for k in range(min(len(path_components), len(namespace_stack)))]

def gen_crumb_tree_menu_items(path, directory):
    for component, name, crumb, title in directory.get_exports():
        if crumb: # skip crumbs and menus if no crumb
            css_class = 'menu_' + component.replace('.', '_')
            yield (path + component, crumb, title, css_class)

def get_crumb_tree():
    """() - > [[(path:str, crumb:str, title:str, css_class:str)]]
    Each list in the result is a menu of exports for a directory on the
    current path with a derived css class name for each menu entry.
    """
    crumb_tree = []
    for path, directory in get_path_directory_list():
        menu = list(gen_crumb_tree_menu_items(path, directory))
        if menu:
            crumb_tree.append(menu)
    return crumb_tree


class Redirector (Directory):
    """
    For use in _q_lookup() implementations that want to consume
    the remaining path components and then issue a redirect.
    """
    def __init__(self, location, permanent=False):
        self.location = location
        self.permanent = permanent
        self.components = []

    def _q_lookup(self, component):
        self.components.append(component)
        return self

    def __call__(self):
        path_info_parts = get_request().get_path_info().split('/')
        while self.components:
            component = self.components.pop()
            assert path_info_parts[-1] == component
            path_info_parts.pop()
        path = '/'.join(path_info_parts)
        return redirect(urljoin(path, self.location), self.permanent)





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/mail/send.py $
$Id: send.py 28696 2006-09-01 11:03:16Z dbinger $
"""
from smtplib import SMTP
from qp.mail.rfc822_mailbox import rfc822_mailbox
from qp.pub.common import get_publisher
import socket

class Email(object):

    max_header_recipients = 10

    def __init__(self):
        self.set_subject('')
        self.set_body('')
        webmaster = self.get_webmaster_address()
        self.set_from_address(webmaster)
        self.set_smtp_sender(webmaster)
        self.set_to([webmaster])
        self.set_cc([])
        self.set_reply_to([])
        self.set_extra_headers([])
        self.set_smtp_recipients([])
        self.sent = False

    def set_subject(self, subject):
        self.subject = str(subject)

    def get_subject(self):
        return self.subject

    def set_body(self, body):
        self.body = str(body)

    def get_body(self):
        return self.body

    def set_from_address(self, address):
        self.from_address = rfc822_mailbox(address)

    def get_from_address(self):
        return self.from_address

    def set_reply_to(self, addresses):
        self.reply_to = self.mailboxes(addresses)

    def get_reply_to(self):
        return self.reply_to

    def set_smtp_sender(self, address):
        self.smtp_sender = rfc822_mailbox(address)

    def get_smtp_sender(self):
        return self.smtp_sender

    def set_to(self, addresses):
        self.to = self.mailboxes(addresses)

    def get_to(self):
        return self.to

    def set_cc(self, addresses):
        self.cc = self.mailboxes(addresses)

    def get_cc(self):
        return self.cc

    def set_extra_headers(self, headers):
        self.extra_headers = [str(header) for header in headers]

    def get_extra_headers(self):
        return self.extra_headers

    def set_smtp_recipients(self, addresses):
        self.smtp_recipients = self.mailboxes(addresses)

    def get_smtp_recipients(self):
        if self.smtp_recipients:
            return self.smtp_recipients
        else:
            return self.get_to() + self.get_cc()

    def get_bulk_headers(self):
        to = self.get_to()
        if len(to) == 1 and to == [self.get_webmaster_address()]:
            return []
        else:
            return ["Precedence: junk", "X-No-Archive: yes"]

    def get_smtp_server(self):
        return get_publisher().get_smtp_server()

    def format_headers(self):
        headers = ["From: %s" % self.get_from_address().format(),
                   "Subject: %s" % self.get_subject()]
        def address_headers(field_name, addresses):
            if not addresses:
                return
            if len(addresses) == 1:
                headers.append("%s: %s" % (field_name, addresses[0].format()))
            elif len(addresses) <= self.max_header_recipients:
                headers.append("%s: %s," % (field_name, addresses[0].format()))
                for address in addresses[1:-1]:
                    headers.append("    %s," % address.format())
                headers.append("    %s" % addresses[-1].format())
            else:
                headers.append(
                    "%s: (long recipient list suppressed) : ;" % field_name)
        address_headers("To", self.get_to())
        address_headers("Cc", self.get_cc())
        address_headers("Reply-To", self.get_reply_to())
        headers += self.get_extra_headers()
        headers += self.get_bulk_headers()
        return '\n'.join(headers) + '\n\n'

    def get_smtp_sender(self):
        return self.smtp_sender

    def set_smtp_sender(self, address):
        self.smtp_sender = rfc822_mailbox(address)

    def is_email_enabled(self):
        return get_publisher().is_email_enabled()

    def get_webmaster_address(self):
        return rfc822_mailbox(get_publisher().get_webmaster_address())

    def get_debug_address(self):
        return rfc822_mailbox(get_publisher().get_debug_address())

    def mailboxes(self, addresses):
        return [rfc822_mailbox(address) for address in addresses if address]

    def send(self):
        assert not self.sent # Make sure duplicates are not sent.
        if not self.is_email_enabled():
            return False
        debug_address = self.get_debug_address()
        smtp_recipients = self.get_smtp_recipients()
        if debug_address:
            recipients = [debug_address.get_addr_spec()]
            message = self.format_headers()
            message += ('[debug mode, message actually sent to %s]\n' %
                        debug_address.format())
            if self.get_smtp_recipients():
                message += ('[original SMTP recipients: %s]\n' %
                            ', '.join([address.format()
                                       for address in smtp_recipients]))
            message += '-' * 72 + '\n'
            message += self.get_body()
        else:
            recipients = [address.get_addr_spec()
                          for address in smtp_recipients]
            message = self.format_headers() + self.get_body()
        sender = self.get_smtp_sender().get_addr_spec()
        server = self.get_smtp_server()
        assert server
        assert sender
        assert recipients
        assert message
        try:
            smtp = SMTP(server)
            smtp.sendmail(sender, recipients, message)
            smtp.quit()
        except socket.error, e:
            print "Email.send() failed: %s" % e
            return False
        self.sent = True
        return True

def sendmail(subject, msg_body, to_addrs, from_addr=None, reply_to=None,
             smtp_sender=None, cc_addrs=None, extra_headers=None,
             mime_type="text/plain", charset='utf8',
             smtp_recipients=None):
    email = Email()
    email.set_subject(subject)
    email.set_body(msg_body)
    email.set_to(to_addrs)
    if from_addr:
        email.set_from_address(from_addr)
    if reply_to:
        email.set_reply_to(reply_to)
    if smtp_sender:
        email.set_smtp_sender(smtp_sender)
    if cc_addrs:
        email.set_cc(cc_addrs)
    extra_headers = extra_headers or []
    extra_headers.append("Content-Type: %s; charset=%s" % (mime_type, charset))
    email.set_extra_headers(extra_headers)
    if smtp_sender:
        email.set_smtp_sender(smtp_sender)
    if smtp_recipients:
        email.set_smtp_recipients(smtp_recipients)
    return email.send()





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/demo/hello/slash.py $
$Id: slash.py 28044 2006-03-09 16:35:08Z dbinger $
"""
from qp.pub.publish import Publisher

class SitePublisher (Publisher):

    configuration = dict(
        http_address=('', 8000),
        as_https_address=('localhost', 9000),
        https_address=('localhost', 10000),
        )

    def process_hit(self, hit):
        hit.get_response().set_body('Hello')




#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/test/utest_tz.py $
$Id: utest_tz.py 28672 2006-08-30 00:53:48Z dbinger $
"""
from datetime import datetime, timedelta
from qp.lib import tz
from sancho.utest import UTest


class Test(UTest):

    def a(self):
        eastern = datetime(
            year=2000, month=12, day=1, hour=10,
            tzinfo=tz.Eastern)
        central = eastern.astimezone(tz.Central)
        mountain = eastern.astimezone(tz.Mountain)
        pacific = eastern.astimezone(tz.Pacific)
        assert eastern.hour == central.hour + 1
        assert eastern.hour == mountain.hour + 2
        assert eastern.hour == pacific.hour + 3
        utc = eastern.astimezone(tz.UTC)
        utc_plus_1 = eastern.astimezone(tz.UTC_plus_1)
        assert utc.hour == eastern.hour + 5
        est = datetime(
            year=2000, month=6, day=1, hour=18,
            tzinfo=tz.Eastern)
        utc2 = est.astimezone(tz.UTC)
        assert utc2.hour == est.hour + 4
        eastern_now = tz.Eastern.now()
        central_now = tz.Eastern.now()
        print  central_now - eastern_now < timedelta(seconds=1)

if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/spec.py $
$Id: spec.py 28959 2006-10-23 13:55:40Z dbinger $

This module provides tools for matching values to specs.

In particular, it provides the functions
  match(),
  require()

This module also provides tools for constructing specs,
using them to specify and validate instance attributes.
"""

import re
from datetime import datetime
from types import FunctionType, MethodType

def format_spec(spec):
    """
    Returns the canonical string representation of the spec.
    """
    if spec is None:
        return 'None'
    if type(spec) is tuple:
        return '(%s)' % ', '.join(map(format_spec, spec))
    if type(spec) is list:
        return '[%s]' % ', '.join(map(format_spec, spec))
    if type(spec) is dict:
        return ('{%s}' %
                ', '.join(["%s: %s" % (format_spec(key_type),
                                       format_spec(val_type)) #nocover
                           for key_type, val_type in spec.items()]))
    if hasattr(spec, '__name__'):
        return spec.__name__
    if isinstance(spec, basestring):
        return repr(spec)
    return str(spec)

class SpecOperator:
    """
    The superclass of all of the spec operators.
    Constructors of subclasses should assign to self.args
    a tuple of the arguments to the constructor.
    """

    def get_args(self):
        return self.args

    def format_args(self):
        return ', '.join(map(format_spec, self.get_args()))

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.format_args())

    def explain_difference(self, value):
        return format_expected_got(self, value)

def match(value, spec):
    """
    Return True or False depending on whether or not value matches
    the given type specification.  Here are the available type
    specifications:

    Type Spec                Matches

    a type or class         any instance of the type

    a callable              anything for which <spec>(value) is true,
                            so you can do arbitrary checking.

    a list (or tuple)       any list (or tuple) whose elements all match the
    with length 1           enclosed type specification.

    a list (or tuple)       any list (tuple) whose sequence of elements match
    with length > 1,        the sequence of corresponding type specifications,
    not containing None,
    or any string or numeric
    or bool literal.

    a tuple with            any value that matches any of the items in the tuple.
    length > 1
    containing None,
    or any string or
    numeric
    literal.
    
    a set                   any set whose elements match any of the elements in
                            the type specification.

    a dict                  any dict for which every item matches some item in
                            this dictionary.

    other                   any value for which value == other.

    """
    if isinstance(spec, (FunctionType, SpecOperator, Specification)):
        return spec(value)
    spec_type = type(spec)
    if spec_type is tuple and len(spec) > 1:
        # Possible disjunction.
        for element in spec:
            if (element is None or
                type(element) in (float, int, str, unicode, bool)):
                # Definite disjunction.
                for element in spec:
                    if match(value, element):
                        return True
                return False
    if spec_type in (list, tuple):
        if type(value) is not spec_type:
            return False
        if len(spec) == 1:
            element_type = spec[0]
            for element in value:
                if not match(element, element_type):
                    return False
            return True
        else:
            if len(value) != len(spec):
                return False
            for element, element_type in zip(value, spec):
                if not match(element, element_type):
                    return False
            return True
    if spec_type is dict:
        if type(value) is not dict:
            return False
        for key, val in value.items():
            for key_type, value_type in spec.items():
                if (match(key, key_type) and
                    match(val, value_type)):
                    break
            else:
                return False
        return True
    if spec_type is set:
        for element in value:
            for element_type in spec:
                if match(element, element_type):
                    break
            else:
                return False
        return True
    try:
        return isinstance(value, spec)
    except TypeError:
        pass
    if spec_type is type(value):
        return value == spec
    if callable(spec):
        return spec(value)
    try:
        return value == spec
    except (TypeError, ValueError, AssertionError):
        pass
    return False

def format_expected_got(spec, value):
    return ('\n  Expected: %s\n'
            '  Got: %r\n') % (format_spec(spec), value)

def require(value, spec, message=None):
    if not match(value, spec):
        if isinstance(spec, SpecOperator):
            error = spec.explain_difference(value)
        else:
            error = format_expected_got(spec, value)
        if message:
            error = '(%s)%s' % (message, error)
        raise TypeError(error)

class Anything (object):
    """
    The universal spec.
    """
    def __str__(self):
        return "anything"

    def __call__(self, value):
        return True

anything = Anything()

class sequence (SpecOperator):

    """
    Use this to specify a sequence.
    """

    def __init__(self, element_spec=anything, container_spec=anything):
        self.container_spec = container_spec
        self.element_spec = element_spec
        self.args = (element_spec, container_spec)

    def __call__(self, value):
        if not match(value, self.container_spec):
            return False
        element_spec = self.element_spec
        for element in value:
            if not match(element, element_spec):
                return False
        return True


class mapping (SpecOperator):

    """
    Use this to specify a mapping.
    """

    def __init__(self, dict_spec=anything, container_spec=anything):
        self.dict_spec = dict_spec
        self.container_spec = container_spec
        self.args =  (self.dict_spec, container_spec)

    def __call__(self, value):
        if not match(value, self.container_spec):
            return False
        try:
            dict_value = dict(value.items())
        except AttributeError:
            return False
        if not match(dict_value, self.dict_spec):
            return False
        return True

class subclass (SpecOperator):

    """
    Use this to specify a subclass of a given class.
    """

    def __init__(self, klass):
        self.klass = klass
        self.args = (klass,)

    def __call__(self, value):
        try:
            return issubclass(value, self.klass)
        except TypeError:
            return False


class instance (SpecOperator):

    """
    Use this to specify an instance of a class with a given name
    when the class itself is not available.
    """

    def __init__(self, klass_name):
        self.klass_name = klass_name
        self.args = (klass_name,)

    def __call__(self, value):
        if hasattr(value, '__class__'):
            classes = [value.__class__]
            while classes:
                if classes[0].__name__ == self.klass_name:
                    return True
                classes = classes[1:] + list(classes[0].__bases__)
        return False


class ConnectiveSpecOperator (SpecOperator):

    def __init__(self, *specs):
        self.specs = specs
        self.args = self.specs


class no (ConnectiveSpecOperator):

    """
    Use this to specify a negation of specs.
    Example: no(None)
    """

    def __call__(self, value):
        for spec in self.specs:
            if match(value, spec):
                return False
        return True


class both (ConnectiveSpecOperator):

    """
    Use this to specify a conjunction of specs.
    Example:
      both([int], length(3)) specifies values that are lists of 3 ints.
    """

    def __call__(self, value):
        for spec in self.specs:
            if not match(value, spec):
                return False
        return True

    def explain_difference(self, value):
        error = format_expected_got(self, value)
        for spec in self.specs:
            if not match(value, spec):
                error += "  (which does not match %s)" % format_spec(spec)
                break
        return error

class either (ConnectiveSpecOperator):

    """
    Use this to specify a disjunction of specs.
    Examples:
      either(int, str) specifies values that are ints or strs.
    """

    def __call__(self, value):
        for spec in self.specs:
            if match(value, spec):
                return True
        return False


class proper (both):

    """
    Use this to specify an instance with attributes that agree with the
    attribute specifications for this class.
    If arguments are provided, they are additional specs which the instance
    must satisfy.
    """

    def __call__(self, value):
        return both.__call__(self, value) and get_spec_problems(value) == []


class interval (SpecOperator):

    def __init__(self, min=anything, max=anything):
        self.min = min
        self.max = max

    def format_args(self):
        if self.max is anything:
            if self.min is anything:
                return ''
            return format_spec(self.min)
        else:
            return '%s, %s' % (format_spec(self.min), format_spec(self.max))

    def __call__(self, value):
        if self.min not  in (None, anything):
            if self.max is anything:
                return value >= self.min
            elif value < self.min:
                return False
        if self.max not  in (None, anything):
            if self.min is anything:
                return value <= self.max
            elif value > self.max:
                return False
        return True


class length (interval):

    """
    Use this to specify specs by ranges of length.
    Examples:
      length(3, None) specifies values of length 3 or more.
      length(3) specifies values of length 3.
    """

    def __call__(self, value):
        return interval.__call__(self, len(value))


class eq (SpecOperator):

    """
    Use this to specify specs exact value.
    Examples:
      eq(SpecOperator) matches the SpecOperator class, but nothing else.

    """

    def __init__(self, value):
        self.value = value

    def format_args(self):
        return repr(self.value)

    def __call__(self, value):
        return value is self.value


class equal (SpecOperator):

    """
    Use this to specify specs exact value.
    Examples:
      eq([2,3]) matches any list [2,3]

    """

    def __init__(self, value):
        self.value = value

    def format_args(self):
        return repr(self.value)

    def __call__(self, value):
        return value == self.value

boolean = either(True, False, 0, 1)


class charset (SpecOperator):

    def __init__(self, charset):
        self.charset = charset

    def __str__(self):
        return self.charset

    def __call__(self, value):
        if isinstance(value, unicode):
            try:
                value.encode(self.charset)
            except UnicodeEncodeError:
                return False
        elif isinstance(value, str):
            try:
                unicode(value, self.charset)
            except UnicodeDecodeError:
                return False
        else:
            return False
        return True


ascii = charset('ascii')


class String (SpecOperator):

    """
    This is the common string class, it ascii str instances and
    all unicode instances.
    """

    def __str__(self):
        return 'string'

    def __call__(self, value):
        if isinstance(value, unicode):
            return True
        elif isinstance(value, str):
            try:
                unicode(value, 'ascii')
                return True
            except UnicodeDecodeError:
                return False
        return False

string = String()


class pattern (SpecOperator):
    """
    Matches basestrings that match the regular expression.
    """
    def __init__(self, pattern):
        self.pattern = re.compile(pattern)
        self.args = (pattern,)

    def __call__(self, value):
        try:
            return bool(self.pattern.match(value))
        except TypeError:
            return False

identifier_pattern = pattern('[a-zA-Z_][a-zA-Z0-9_]*$')

class with_attribute (SpecOperator):
    """
    Matches values with attributes matching given specs.
    A keyword argument gives the spec for each attribute.
    """
    def __init__(self, **attribute_specs):
        self.attribute_specs = attribute_specs

    def __call__(self, value):
        for attribute, spec in self.attribute_specs.items():
            if (not hasattr(value, attribute) or
                not match(getattr(value, attribute), spec)):
                return False
        return True

    def format_args(self):
        return ', '.join(['%s=%s' % (name, format_spec(spec))
                          for name, spec in sorted(self.attribute_specs.items())])


datetime_without_tz = both(datetime, with_attribute(tzinfo=None))
datetime_with_tz = both(datetime, with_attribute(tzinfo=no(None)))

class Specification:
    """
    Instance attributes:
      doc : string
      spec : anything
    """

    def __init__(self, spec, doc=''):
        self.doc = doc
        self.spec = spec

    def __call__(self, value):
        return match(value, self.spec)

    def __str__(self):
        """-> str
        Return a formatted string describing this spec and corresponding doc.
        """
        doc = '\n  '.join([line.strip()
                           for line in self.doc.split('\n')])
        if doc:
            return self.format_spec() + '\n  ' + doc
        else:
            return self.format_spec()

    def valid(self, value):
        """(value : anything) -> bool
        Does the value match this spec?
        """
        return match(value, self.spec)

    def format_spec(self):
        """() -> str
        Return a string version of the spec.
        """
        return format_spec(self.spec)

def get_spec(klass, name):
    """(klass:anything, name:str) -> anything
    """
    value = getattr(klass, name)
    # functions as specifiers must be unwrapped, since
    # assignment to class variables changes them.
    if type(value) is MethodType: # normal classes
        spec = value.im_func
    else:
        spec = value
    return spec

def spec(spec, *docs):
    """(anything, *str) -> Specification
    An alternative Specification constructor.
    """
    doc = '\n'.join(docs)
    return Specification(spec, doc)

def nspec(spec, *docs):
    """(anything, *str) -> Specification
    Like spec, but this uses a modified Specification (if needed) to make
    sure that None is a valid value.
    For applications that always want None as a valid value, this
    allows us to avoid using the either(None,*) pattern around every spec.
    """
    doc = '\n'.join(docs)
    s = Specification(spec, doc)
    if s.valid(None):
        return s
    if isinstance(spec, either):
        return Specification(either(None, *spec.get_args()), doc)
    return Specification(either(None, spec), doc)

def get_specs(klass):
    """(klass:anything) -> { str : anything }
    Returns a dictionary mapping names to specs that apply to instances of
    klass.  Specs are distinguished by the following naming convention:
    A class attribute whose name ends in '_is' specifies an instance
    attribute without the suffix.
    Example:
    class A:
        color_is = either(str, None)
        def __init__(self, x):
            self.color = x
    """
    specs = {}
    for name in dir(klass):
        if name.endswith('_is'):
            specs[name[:-3]] = get_spec(klass, name)
    return specs

def persistent_vars(obj):
    if hasattr(obj, '__getstate__'):
        return obj.__getstate__() or {}
    else:
        return vars(obj)

def get_spec_problems(x, specs=None):
    """(x:anything, specs:{str:Spec}) -> [basestring]
    Return a list of reasons why the attributes of this
    instance do not match the specs.
    Use provided specs for speed if you already
    have the specs of the class of x.
    """
    reasons = []
    if specs is None:
        specs = get_specs(x.__class__)
    for name, spec in specs.items():
        if not hasattr(x, name):
            reasons.append(('%r.%s: Missing\n'
                            '  Expected: %s\n') % (x, name, format_spec(spec)))
        else:
            value = getattr(x, name)
            if not match(value, spec):
                reasons.append(
                    ('%r.%s:\n' #nocover
                     '  Expected: %s\n'
                     '  Got: %r\n') % ( #nocover
                    x, name, format_spec(spec), value))
    for name, value in persistent_vars(x).iteritems():
        if not name in specs:
            reasons.append('Found non-specified attribute: %r.%s (=%r)' %
                           (x, name, value))
    return reasons

def get_spec_doc(klass):
    """(klass:Specified) -> str
    Return a string containing the docs of all specs of klass.
    """
    spec_items = get_specs(klass).items()
    spec_items.sort()
    return '\n'.join(["%s: %s" % (name, format_spec(spec))
                      for name, spec in spec_items])

def get_spec_report(instances):
    """(instances: [InstanceType]) -> str|None
    Study the instances with respect to their attribute specs.
    The instances must all be of the same class.
    If any instance does not satisfy the specs, return a
    string explaining the problem.
    If all instances satisfy the specs, return a message reporting
    the number of instances verified.
    If the class of the instances has no specifications,
    return a message that says so.
    If there are no instances, return None.
    """
    count = 0
    klass = specs = None
    for instance in instances:
        if klass is None:
            klass = instance.__class__
            specs = get_specs(klass)
            if not specs:
                return "\nunspecified " + klass.__name__
        assert instance.__class__ is klass
        count += 1
        instance_problems = get_spec_problems(instance, specs=specs)
        if instance_problems:
            return "\n%s:\n" % klass.__name__ + "\n".join(instance_problems)
    return klass and "\n%s: %s ok\n" % (klass.__name__ , count)

def specify(obj, **attributes_values):
    """(obj:object, **attributes_values:{str:anything})
    Set object attributes given as keywords.
    Raise a TypeError if any value does match a spec,
    and AttributeError if no corresponding spec is found.
    """
    for attribute, value in attributes_values.items():
        require(value, getattr(obj.__class__, attribute + '_is'))
        setattr(obj, attribute, value)

def init(obj, **attributes_values):
    """(obj:object, **attributes_values:{str:anything})
    Set object attributes given as keywords.
    Raise a TypeError if any value does match a spec,
    and AttributeError if no corresponding spec is found.
    In addition, all other specified attributes are set to None,
    without any type checking.
    """
    for name in get_specs(obj.__class__):
        if not hasattr(obj, name):
            setattr(obj, name, None)
    specify(obj, **attributes_values)

def add_getters(klass):
    """
    Add trivial getter method for each specified attribute
    for which no such method currently exists.
    """
    def add_getter(klass, name):
        accessor_name = 'get_' + name
        if not hasattr(klass, accessor_name):
            def get_specified_attribute(self):
                return getattr(self, name)
            get_specified_attribute.__name__ = accessor_name
            setattr(klass, accessor_name, get_specified_attribute)
    for name in get_specs(klass):
        add_getter(klass, name)

def add_setters(klass):
    """
    Add trivial setter method for each specified attribute
    for which no such method currently exists.
    """
    def add_setter(klass, name):
        setter_name = 'set_' + name
        if not hasattr(klass, setter_name):
            def f(self, value):
                require(value, getattr(klass, name + '_is'))
                setattr(self, name, value)
            f.__name__ = setter_name
            setattr(klass, setter_name, f)
    for name in get_specs(klass):
        add_setter(klass, name)

def add_getters_and_setters(klass):
    add_getters(klass)
    add_setters(klass)

class SlotsFromSpecs (type):
    """
    Classes with this metaclass that do not already have an
    explicit __slots__ attribute will get one that is constructed
    from the attributes (including inherited ones) whose names end 
    with "_is".
    """
    def __new__(self, klass_name, bases, namespace):
        if '__slots__' not in namespace:
            names = set()
            for name in namespace:
                if name.endswith('_is'):
                    names.add(name[:-3])
            for base in bases:
                for name in dir(base):
                    if name.endswith('_is'):
                        names.add(name[:-3])
            namespace['__slots__'] = sorted(names)
        return type.__new__(self, klass_name, bases, namespace)

class Specified (object):
    """
    This mixin class has a metaclass that uses attribute specs to build
    a __slots__ list for the class.
    """
    __metaclass__ = SlotsFromSpecs



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/hub/dispatcher.py $
$Id: dispatcher.py 28406 2006-05-24 16:47:44Z dbinger $

For systems that support passfd, this provides
a pre-forking server that uses file descriptor passing to off-load
requests to child worker processes.

"""
try:
    from qp.hub import passfd
except ImportError:
    passfd = None

import errno
import fcntl
import os
import select
import socket
import stat
import sys

class ModelConnectionHandler(object):
    """
    The create_connection_handler argument to HubServer.__init__() needs
    to act like the constructor of this class: calling it with no arguments
    returns something with a 'handle_connection' method with a signature
    like the one here.
    """
    def __init__(self):
        """This is executed in the child process, when it starts."""

    def handle_connection(self, connection):
        """This does whatever is needed to read a request and respond."""



class HubServer(object):

    banned_msg = (
        'HTTP/1.0 403 Forbidden\r\n'
        'Content-Type: text/plain\r\n\r\n'
        'Your IP address has been banned from accessing this site.\n'
        'Please email the webmaster here if you need the ban removed.')

    def __init__(self, create_connection_handler, max_children=5, banned=None):
        self.create_connection_handler = create_connection_handler
        if passfd:
            self.max_children = max_children
            self.connection_handler = None
        else:
            # We can't pass file descriptors.
            # Fall back to non-forking mode.
            self.max_children = 0
            self.connection_handler = create_connection_handler()
        self.children = {} # { pid : fd }
        self.banned = banned
        self.banned_time = -1
        self.banned_addresses = set()

    def spawn_child(self, conn=None):
        if len(self.children) >= self.max_children:
            return
        parent_fd, child_fd = passfd.socketpair(socket.AF_UNIX,
                                                socket.SOCK_STREAM)
        # make child fd non-blocking
        flags = fcntl.fcntl(child_fd, fcntl.F_GETFL, 0)
        fcntl.fcntl(child_fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
        pid = os.fork()
        if pid == 0:
            if conn:
                conn.close() # in the midst of handling a request, close
                             # the connection in the child
            os.close(child_fd)
            self.serve(parent_fd)
            sys.exit(0)
        else:
            os.close(parent_fd)
            self.children[pid] = child_fd

    def reap_children(self):
        while self.children:
            (pid, status) = os.waitpid(-1, os.WNOHANG)
            if pid <= 0:
                break
            os.close(self.children[pid])
            del self.children[pid]

    def do_stop(self):
        #
        # First close connections to the children, which will cause them
        # to exit after finishing what they are doing.
        #
        for fd in self.children.values():
            os.close(fd)
        #
        # Then do a blocking wait on each until we have cleared the
        # slate.
        #
        for pid in self.children.keys():
            (pid, status) = os.waitpid(pid, 0)
        self.children = {}

    def delegate_request(self, conn):
        """Pass a request fd to a child process to handle.  This method
        blocks if all the children are busy and we have reached the
        max_children limit."""

        # There lots of subtleties here.  First, we can't use the write
        # status of the pipes to the child since select will return true
        # if the buffer is not filled.  Instead, each child writes one
        # byte of data when it is ready for a request.  The normal case
        # is that a child is ready for a request.  We want that case to
        # be fast.  Also, we want to pass requests to the same child if
        # possible.  Finally, we need to gracefully handle children
        # dying at any time.

        if self.banned:
            try:
                mtime = os.stat(self.banned)[stat.ST_MTIME]
            except OSError, e:
                pass
            else:
                if mtime > self.banned_time:
                    self.banned_time = mtime
                    f = open(self.banned)
                    new_banned_addresses = set(
                        [a.strip() for a in f.read().split()])
                    f.close()
                    for address in (self.banned_addresses -
                                    new_banned_addresses):
                        print 'UNBANNED:', address
                    for address in (new_banned_addresses -
                                    self.banned_addresses):
                        print 'BANNED:', address
                    self.banned_addresses = new_banned_addresses
                client_address_port = conn.getpeername()
                if client_address_port[0] in self.banned_addresses:
                    conn.send(self.banned_msg)
                    conn.close()
                    return

        if self.max_children == 0:
            # We are in non-forking mode.  Just handle the request here.
            self.connection_handler.handle_connection(conn)
            return

        # If no children are ready and we haven't reached max_children
        # then we want another child to be started without delay.
        timeout = 0

        while 1:
            try:
                r, w, e = select.select(self.children.values(), [], [], timeout)
            except select.error, e:
                if e[0] == errno.EINTR:  # got a signal, try again
                    continue
                raise
            if r:
                # One or more children look like they are ready.  Sort
                # the file descriptions so that we keep preferring the
                # same child.
                r.sort()
                child_fd = r[0]

                # Try to read the single byte written by the child.
                # This can fail if the child died or the pipe really
                # wasn't ready (select returns a hint only).  The fd has
                # been made non-blocking by spawn_child.  If this fails
                # we fall through to the "reap_children" logic and will
                # retry the select call.
                try:
                    ready_byte = os.read(child_fd, 1)
                    if not ready_byte:
                        raise IOError # child died?
                    assert ready_byte == "1", repr(ready_byte)
                except socket.error, exc:
                    if exc[0]  == errno.EWOULDBLOCK:
                        pass # select was wrong
                    else:
                        raise
                except (OSError, IOError):
                    pass # child died?
                else:
                    # The byte was read okay, now we need to pass the fd
                    # of the request to the child.  This can also fail
                    # if the child died.  Again, if this fails we fall
                    # through to the "reap_children" logic and will
                    # retry the select call.
                    try:
                        passfd.sendfd(child_fd, conn.fileno())
                    except IOError, exc:
                        if exc.errno == errno.EPIPE:
                            pass # broken pipe, child died?
                        else:
                            raise
                    else:
                        # fd was apparently passed okay to the child.
                        # The child could die before completing the
                        # request but that's not our problem anymore.
                        return

            # didn't find any child, check if any died
            self.reap_children()

            # start more children if we haven't met max_children limit
            #conn.close() and not pass conn?
            if len(self.children) < self.max_children:
                self.spawn_child(conn)

            # Start blocking inside select.  We might have reached
            # max_children limit and they are all busy.
            timeout = 2

    def serve(self, parent_fd):
        handler = self.create_connection_handler()
        while 1:
            try:
                os.write(parent_fd, "1") # indicates that child is ready
                fd = passfd.recvfd(parent_fd)
            except (IOError, OSError):
                # parent probably exited  (EPIPE comes thru as OSError)
                raise SystemExit
            conn = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
            # Make sure the socket is blocking.  Apparently, on FreeBSD the
            # socket is non-blocking.  I think that's an OS bug but I don't
            # have the resources to track it down.
            conn.setblocking(1)
            os.close(fd)
            handler.handle_connection(conn)
            conn.close()

    def listen(self, *addresses):
        name = 'QP_HUBSERVER_FILE_DESCRIPTORS'
        sockets = []
        addresses = filter(None, addresses)
        if os.environ.get(name):
            fds = map(int, os.environ.get(name).split(','))
            assert len(fds) == len(addresses)
            sockets = [socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
                for fd in fds]
        if not sockets:
            def get_socket_bound_to_address(address):
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                s.bind(address)
                s.listen(40)
                return s
            sockets = [get_socket_bound_to_address(address)
                for address in addresses]
        os.environ[name] = ','.join([str(s.fileno()) for s in sockets])
        self.listening_sockets = sockets

    def run(self):
        self.spawn_child()
        while 1:
            r, w, e = select.select(self.listening_sockets, [], [], None)
            for s in r:
                try:
                    conn, addr = s.accept()
                    self.delegate_request(conn)
                    conn.close()
                except socket.error, e:
                    if e[0] != errno.EINTR:
                        raise  # something weird






"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/delegation.py $
$Id: delegation.py 28598 2006-08-16 19:16:51Z dbinger $
"""

def delegate(klass, *names):
    """
    The names are strings of the form <attribute_name>.<method_name>.
    For each one, this adds a method to klass, named <method_name>,
    that delegates to the same method of self.<attribute_name>.
    Arguments are passed to the method without change.
    """
    def get_delegated(attribute_name, method_name):
        def delegated(self, *args, **kwargs):
            method = getattr(getattr(self, attribute_name), method_name)
            return method(*args, **kwargs)
        return delegated
    for name in names:
        attribute_name, method_name = name.split('.')
        setattr(klass, method_name, get_delegated(attribute_name, method_name))

def delegate_passing_self(klass, *names):
    """
    The names are strings of the form <attribute_name>.<method_name>.
    For each one, this adds a method to klass, named <method_name>,
    that delegates to the same method of self.<attribute_name>.
    Arguments are passed to the method with "self" added as the first
    argument.
    """
    def get_delegated(attribute_name, method_name):
        def delegated(self, *args, **kwargs):
            method = getattr(getattr(self, attribute_name), method_name)
            return method(self, *args, **kwargs)
        return delegated
    for name in names:
        attribute_name, method_name = name.split('.')
        setattr(klass, method_name, get_delegated(attribute_name, method_name))


#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/bin/qpcensus.py $
$Id: qpcensus.py 28950 2006-10-19 19:08:15Z rmasse $
"""
from cPickle import loads
from datetime import datetime
from durus.connection import Connection
from durus.file_storage import FileStorage
from durus.serialize import unpack_record, split_oids
from qp.lib.site import Site
from qp.lib.spec import get_spec_problems, get_specs, get_spec_doc
from qp.lib.spec import persistent_vars
import sys


class ObjectLocator:

    def __init__(self, connection, reference_index):
        self.connection = connection
        self.reference_index = reference_index

    def where_is(self, obj):
        root = self.connection.get_root()
        def locate(obj, seen):
            if len(seen) > 50:
                return '...'
            if obj is root:
                return 'root'
            oid = obj._p_oid
            for referring_oid in self.reference_index.get(oid, []):
                if referring_oid not in seen:
                    seen.add(referring_oid)
                    break
            else:
                return repr(obj) # give up
            referring = self.connection.get(referring_oid)
            ref = locate(referring, seen)
            for key, value in persistent_vars(referring).iteritems():
                if obj is value:
                    return "%s.%s" % (ref, key)
                try:
                    for k, v in value.iteritems():
                        if obj is v:
                            return "%s[%r]" % (ref, k)
                except AttributeError:
                    pass
                try:
                    for k, v in enumerate(value):
                        if obj is v:
                            return "%s[%r]" % (ref, k)
                except TypeError:
                    pass
            return '%s ... ' % ref
        return locate(obj, set())


def format_specs(klass):
    if klass.__bases__:
        bases = '(%s):' % ', '.join(
            [b.__name__ for b in klass.__bases__])
    else:
        bases = ":"
    s = "class %s%s\n    " % (klass.__name__, bases)
    s += get_spec_doc(klass).replace('\n', '\n    ') or 'pass'
    s += "\n\n"
    return s

def check_durus(file, schema, klasses, show_all, show_max, show_path):
    storage = FileStorage(file, readonly=True)
    connection = Connection(storage, cache_size=100)
    if schema:
        schema_output = open(schema, 'w')
    problem_record = {}
    specs = {}
    counts = {}
    size_min = {}
    size_max = {}
    maximal = {}
    size_sum = {}
    refs = {}
    instance_count = 0
    start = datetime.now()
    for oid, record in storage.gen_oid_record():
        obj = connection.cache.get(oid)
        record_oid, data, refdata = unpack_record(record)
        if show_path:
            for refoid in split_oids(refdata):
                refs.setdefault(refoid, []).append(oid)
        if obj is None:
            klass = loads(data)
            obj = connection.cache.get_instance(oid, klass, connection)
        if obj._p_is_ghost():
            state = connection.reader.get_state(data, load=True)
            obj.__setstate__(state)
            obj._p_set_status_saved()
        size = len(record)
        klass = obj.__class__
        if klass not in counts:
            counts[klass] = 1
            size_min[klass] = size
            size_max[klass] = size
            size_sum[klass] = size
            maximal[klass] = obj
            specs[klass] = get_specs(obj)
            if schema:
                schema_output.write(format_specs(klass))
        else:
            counts[klass] += 1
            size_min[klass] = min(size, size_min[klass])
            if size > size_max[klass]:
                maximal[klass] = obj
                size_max[klass] = size
            size_sum[klass] += size
        if show_all or klass not in problem_record:
            problems = get_spec_problems(obj, specs=specs[klass])
            if problems:
                problem_record.setdefault(klass, []).append((obj, problems))
        obj._p_set_status_ghost()
        instance_count += 1
        if instance_count % 10000 == 0:
            sys.stdout.write('.')
            sys.stdout.flush()
            connection.abort()
    end = datetime.now()
    print
    print file
    print end
    seconds = (end - start).seconds
    if seconds > 60:
        minutes, seconds = divmod(seconds, 60)
        fmt = "%sm, %ss" % (minutes, seconds)
    else:
        fmt = "%ss" % seconds
    print "Scanned %s instances in in %s." % (instance_count, fmt)
    klasses = sorted(counts, key=lambda x: x.__name__)
    if show_path:
        where_is = ObjectLocator(connection, refs).where_is
    else:
        def where_is(obj):
            return obj._p_format_oid()
    for klass in klasses:
        if klass in problem_record:
            for obj, problem in problem_record[klass]:
                print
                print "Problem with instance of %s:" % (klass.__name__)
                print ' ', where_is(obj)
                for line in problem:
                    if len(line) > 200:
                        line = line[:200] + ' ...'
                    print ' ', line
    print
    print "%7s %-35s %7s %7s %7s %12s" % (
        "Count",
        "Class",
        "Min",
        "Max",
        "Ave",
        "Sum")
    for klass in klasses:
        print "%7s %-35s %7s %7s %7s %12s" % (
            counts[klass],
            klass.__name__,
            size_min[klass],
            size_max[klass],
            size_sum[klass] / counts[klass],
            size_sum[klass])
    if show_max:
        print
        print 'Maximal Size Instances:'
        for klass in klasses:
            print klass.__name__
            print ' ', where_is(maximal[klass])


def check_durus_main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.set_description("Searches for spec anomolies in a durus database.")
    parser.add_option(
        '-s', '--site', dest='site', default=[], action='append',
        help=("Check only the named site.  Default is all sites."))
    parser.add_option(
        '--file', dest="file", default=None,
        help="If this is given, check this file only.")
    parser.add_option(
        '--schema', dest="schema", default=None,
        help="If given, a schema will be written to this file")
    parser.add_option(
        '--class', dest="classes", default=[], action="append",
        help="If given, limit study to this class.")
    parser.add_option(
        '--all', dest="all", action='store_true',
        help="If given, all errors are shown not just the first in each class.")
    parser.add_option(
        '--showmax', dest="show_max", action='store_true',
        help=("If given, an instance of maximal size is identified for"
        " each class."))
    parser.add_option(
        '--showpath', dest="show_path", action='store_true',
        help="If given, work harder to show a path to instances.")
    (options, args) = parser.parse_args()
    if options.file:
        check_durus(options.file, options.schema, options.classes, options.all,
        options.show_max, options.show_path)
    else:
        for name, site in Site.get_sites().items():
            if options.site and name not in options.site:
                continue
            print "%s:" % name
            check_durus(site.get_durus_file(), options.schema,
                        options.classes, options.all, options.show_max, 
                        options.show_path)


if __name__ == '__main__':
    check_durus_main()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/fill/test/utest_form.py $
$Id: utest_form.py 28953 2006-10-20 16:07:28Z dbinger $
"""
from qp.fill.form import Form, get_javascript_code
from qp.fill.widget import MultipleSelectWidget
from qp.pub.common import clear_publisher
from qp.pub.hit import Hit
from qp.pub.publish import DurusPublisher
from sancho.utest import UTest, raises

class FormTest(UTest):

    def _pre(self):
        clear_publisher()

    def _post(self):
        clear_publisher()

    def a(self):
        publisher = DurusPublisher()
        hit = Hit(None, {})
        publisher.process_hit(hit)
        form = Form()
        form.add_submit('a')
        form.add_reset('b')
        form.add_hidden('c')
        form.add_string('d')
        form.add_text('e')
        form.add_checkbox('f')
        form.add_single_select('g', options=[0])
        form.add_multiple_select('h', options=[0])
        form.add_radiobuttons('i', options=[0])
        form.add_float('j')
        form.add_int('k')
        form.render()
        assert not form.is_submitted()

    def b(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        publisher.process_hit(hit)
        form = Form()
        form.add_submit('a')
        form.add_string('d')
        assert form.is_submitted()
        form.render()

    def c(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        publisher.process_hit(hit)
        form = Form(method="get")
        form.add_submit('a')
        form.add_string('d')
        assert form.is_submitted()
        form.render()

    def d(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="POST"))
        publisher.process_hit(hit)
        form = Form(id='ok')
        form.add_submit('a')
        form.add_string('d')
        form.set_error('a', 'err')
        assert not form.is_submitted()
        assert 'id="ok"' in form.render()

    def e(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        hit.get_request().process_inputs()
        publisher.set_hit(hit)
        form = Form(method="get")
        form.add_submit('a')
        form.add_string('d')
        form.has_key('d')
        assert form.get('d') == '1'
        assert form['d'] == '1'
        assert form.is_submitted()
        get_javascript_code()['a'] = 'javascript'
        assert 'javascript' in str(form.render())

    def f(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        hit.get_request().process_inputs()
        publisher.set_hit(hit)
        form = Form(method="get")
        form.add(MultipleSelectWidget, 'a', options=[range(10)])
        class FakePersistent(object): pass
        a = FakePersistent()
        b = FakePersistent()
        a._p_oid = '11111111'
        b._p_oid = '22222222'
        FakePersistent.__str__ = lambda x: 'ok'
        form.add(MultipleSelectWidget, 'b', options = [a, b])
        del a._p_oid
        raises(AssertionError, # because descriptions are not unique.
               form.add, MultipleSelectWidget, 'c', options = [a, b])

    def g(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        hit.get_request().process_inputs()
        publisher.set_hit(hit)
        form = Form(method="get")
        form.add(MultipleSelectWidget, 'a', options=[range(10)])
        form.add_string('d')
        assert not form.get_widget('a').is_submitted()
        assert form.get_widget('d').is_submitted()
        assert form.is_submitted()
        assert form.get_widget('a').is_submitted()

if __name__ == '__main__':
    FormTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/mail/test/utest_send.py $
$Id: utest_send.py 26833 2005-05-23 00:50:41Z dbinger $
"""
from os import getuid, environ
from pwd import getpwuid
from sancho.utest import UTest
from socket import getfqdn
from qp.mail.send import Email
from qp.pub.common import clear_publisher, get_publisher
from qp.pub.publish import Publisher


class Test(UTest):

    def _pre(self):
        Publisher()

    def _post(self):
        clear_publisher()

    def a(self):
        e = Email()
        webmaster = get_publisher().get_webmaster_address()
        assert e.format_headers() == ('From: %(webmaster)s\n'
                                      'Subject: \n'
                                      'To: %(webmaster)s\n\n') % locals()
        e.set_to(['al', 'bo', 'ci'])
        e.set_subject('the subject')
        assert e.format_headers() == ('From: %(webmaster)s\n'
                                      'Subject: the subject\n'
                                      'To: al,\n'
                                      '    bo,\n'
                                      '    ci\n'
                                      'Precedence: junk\n'
                                      'X-No-Archive: yes\n\n') % locals()
        e.set_extra_headers(['Spam: yes', 'Ham: no'])
        e.set_reply_to([webmaster])
        assert e.format_headers() == ('From: %(webmaster)s\n'
                                      'Subject: the subject\n'
                                      'To: al,\n'
                                      '    bo,\n'
                                      '    ci\n'
                                      'Reply-To: %(webmaster)s\n'
                                      'Spam: yes\n'
                                      'Ham: no\n'
                                      'Precedence: junk\n'
                                      'X-No-Archive: yes\n\n') % locals()

    def b(self):
        e = Email()
        e.set_to([str(x) for x in range(30)])
        e.set_cc('a')
        webmaster = get_publisher().get_webmaster_address()
        assert e.format_headers() == ('From: %(webmaster)s\n'
                                      'Subject: \n'
                                      'To: (long recipient list suppressed) : ;\n'
                                      'Cc: a\n'
                                      'Precedence: junk\n'
                                      'X-No-Archive: yes\n\n') % locals()
    def c(self):
        user = "%s@%s" % (getpwuid(getuid()).pw_name, getfqdn())
        e = Email()
        e.set_to([user])
        assert get_publisher().is_email_enabled() == False
        if not environ.get('TESTEMAIL'):
            return
        assert e.send() == False
        def is_email_enabled():
            return True
        get_publisher().is_email_enabled = is_email_enabled
        e.set_body('This was sent when %s ran utest_send.py' % user)
        e.set_subject('utest_send.py')
        assert e.send() == True
        def get_debug_address():
            return None
        get_publisher().get_debug_address = get_debug_address
        f = Email()
        f.set_to([user])
        f.set_subject('utest_send.py')
        f.set_body('This was also sent when %s ran utest_send.py '
                   '\n(with a debug_address).\n' %
                   user)
        assert f.send() == True

if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/mail/test/utest_rfc822_mailbox.py $
$Id: utest_rfc822_mailbox.py 26815 2005-05-17 11:07:27Z dbinger $
"""
from sancho.utest import UTest, raises
from qp.mail.rfc822_mailbox import rfc822_mailbox


class Test(UTest):

    def a(self):
        assert rfc822_mailbox(None) is None
        result = rfc822_mailbox('me')
        assert result.format() == 'me'
        assert result is rfc822_mailbox(result)
        assert raises(TypeError, rfc822_mailbox, 1)
        assert rfc822_mailbox(('a@b', 'bo')).format() == 'bo <a@b>'
        assert rfc822_mailbox(('a@b', 1)).format() == '1 <a@b>'
        assert raises(TypeError, rfc822_mailbox, (1, 2, 3))
        assert rfc822_mailbox(('a@b', 'b "bo" ba')).format() == (
            r'"b \"bo\" ba" <a@b>')

if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/user.py $
$Id: user.py 28414 2006-05-25 19:37:53Z dbinger $
"""
from durus.persistent import Persistent
from durus.persistent_dict import PersistentDict
from durus.persistent_set import PersistentSet
from md5 import md5
from qp.lib.spec import require, specify, spec, add_getters, sequence, either
from qp.lib.spec import pattern
from qp.pub.common import get_publisher
from qp.lib.util import randbytes

class Permissions (PersistentDict):

    data_is = {basestring:sequence(either(Persistent, True), PersistentSet)}

    def grant(self, permission, granter):
        require(permission, basestring)
        require(granter, either(Persistent, True))
        if permission not in self:
            self[permission] = PersistentSet([granter])
        else:
            self[permission].add(granter)

    def ungrant(self, permission, granter):
        require(permission, basestring)
        require(granter, either(Persistent, True))
        if self.is_granted(permission, granter):
            self.data[permission].remove(granter)
            if len(self.data[permission]) == 0:
                del self.data[permission]

    def is_granted(self, permission, granter):
        return granter in self.get(permission, [])


def compute_digest(*args):
    return md5(":".join([str(x) for x in args])).hexdigest()


class Digester (Persistent):

    digests_is = {basestring:basestring}

    def __init__(self):
        self.digests = {}

    def set_digest(self, realm, digest):
        self._p_note_change()
        self.digests[realm] = digest

    def get_digest(self, realm):
        return self.digests.get(realm)

    def get_realms(self):
        return self.digests.keys()

    def remove_digest(self, realm):
        self._p_note_change()
        del self.digests[realm]

    def discard_digest(self, realm):
        if realm in self.digests:
            self.remove_digest(realm)


class TokenSet(Persistent):
    """
    A set of randomly generated tokens that have been delivered with forms.
    These are used to make sure that forms can't be replayed.
    """
    tokens_is = spec(
        [str],
        "The tokens that have been delivered with forms recently.")

    def __init__(self):
        self.clear()

    def clear(self):
        self.tokens = []

    def __contains__(self, token):
        return token in self.tokens

    def remove(self, token):
        self._p_note_change()
        self.tokens.remove(token)

    def discard(self, token):
        if token in self.tokens:
            self.remove(token)

    def new_token(self, bytes=8):
        token = randbytes(bytes)
        self.tokens = ([token] + self.tokens)[:16]
        return token


class User (Persistent):

    id_is = spec(
        pattern('^[-A-Za-z0-9_@.]*$'),
        "unique among users here")
    digester_is = spec(
        Digester,
        "holds password hashes")
    permissions_is = spec(
        Permissions,
        "Records permissions granted.")
    tokens_is = spec(
        TokenSet,
        "a nonce dealer")
    email_is = spec(
        (None, pattern("^.+@.+\..{2,4}$")),
        "User's email address")

    def __init__(self, user_id):
        specify(self,
                digester=Digester(),
                permissions=Permissions(),
                tokens=TokenSet(),
                email=None)
        if user_id is None:
            self.id = None
        else:
            specify(self, id=user_id)

    def __nonzero__(self):
        """The null user is the one with id == ''.
        """
        return self.id != ''

    def set_password(self, password, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        digester = self.get_digester()
        if not password:
            digester.discard_digest(realm)
        else:
            digester.set_digest(
                realm, compute_digest(self.get_id(), realm, password))

    def has_password(self, password, realm=None):
        if realm is None:
            realm = get_publisher().get_site().get_name()
        digester = self.get_digester()
        digest = digester.get_digest(realm)
        return bool(
            digest and
            digest == compute_digest(self.get_id(), realm, password))

    def is_granted(self, permission, other=True):
        """(permission:basestring, other:either(Persistent,True)=True) -> bool
        Does this user have the `permission` granted from the `other`?
        """
        return self.permissions.is_granted(permission, other)

    def is_admin(self):
        """() -> bool
        Has the root object granted 'administrator' to this user?
        """
        return self.permissions.is_granted('administrator', True)

add_getters(User)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/hub/scgi_util.py $
$Id: scgi_util.py 28403 2006-05-24 14:15:41Z dbinger $
"""
import socket

def scgi_request(content, **headers):
    """(content:str, **headers) -> str
    This is intended for testing purposes, for use by send_scgi() (below).
    Returns an scgi-formatted request string using the given content and headers.
    """
    headers = '\0'.join(
        ['%s\0%s' % item for item in
         [('CONTENT_LENGTH', len(content)), ('SCGI', 1)] + headers.items()])
    return "%s:%s\0,%s" % (len(headers)+1, headers, content)

def send_scgi(port, host='localhost', content='',
              PATH_INFO='/', SCRIPT_NAME='', **kwargs):
    """
    This function *only* for testing the response of a running scgi server.
    It sends a request in scgi format and returns the beginning part of the
    response.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    if content:
        method = "POST"
    else:
        method = "GET"
    s.send(scgi_request(content,
                        REQUEST_METHOD=method,
                        PATH_INFO=PATH_INFO,
                        SCRIPT_NAME=SCRIPT_NAME,
                        **kwargs))
    result = s.recv(1000000)
    s.close()
    return result

if __name__ == '__main__':
    import sys
    if len(sys.argv) < 2:
        print '%s <port> [<PATH_INFO>]' % sys.argv[0]
    elif len(sys.argv) == 2:
        path = '/'
    else:
        path = sys.argv[2]
    port = int(sys.argv[1])
    print send_scgi(port, PATH_INFO=path)



from os.path import expanduser, expandvars, isdir
sites_path = ["${QP_SITES}", "~/qp_sites", "~/.qp_sites", '/var/qp_sites']
sites_path += __path__

__path__ = []
for raw_path in sites_path:
    path = expanduser(expandvars(raw_path))
    if isdir(path):
        __path__.append(path)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/hub/web.py $
$Id: web.py 28677 2006-08-31 15:48:07Z dbinger $

Provides a HubServer-based HTTP/SCGI server.
"""
from BaseHTTPServer import BaseHTTPRequestHandler
from qp.hub.dispatcher import HubServer
from qp.lib.site import Site
from qp.pub.common import get_publisher, clear_publisher
from urllib import unquote
from signal import signal, SIGHUP
import os
import qp
import socket
import sys

class HTTPRequestHandler(BaseHTTPRequestHandler):

    server_version = "qp/%s" % qp.__version__

    def __init__(self, connection, address, info, **kwargs):
        self.required_cgi_environment = dict(
            SERVER_SOFTWARE=self.server_version,
            GATEWAY_INTERFACE='CGI/1.1')
        self.required_cgi_environment.update(kwargs)
        BaseHTTPRequestHandler.__init__(
            self, connection, address, info)

    def get_cgi_env(self, method):
        env = dict(
            SERVER_NAME=self.server.server_name,
            SERVER_PROTOCOL=self.protocol_version,
            SERVER_PORT=str(self.server.server_port),
            REQUEST_METHOD=method,
            REMOTE_ADDR=self.client_address[0],
            REMOTE_PORT=self.client_address[1])
        if '?' in self.path:
            env['PATH_INFO'], env['QUERY_STRING'] = self.path.split('?', 1)
        else:
            env['PATH_INFO'] = self.path
        env['PATH_INFO'] = unquote(env['PATH_INFO'])
        script_name = self.required_cgi_environment.get('SCRIPT_NAME', '')
        if script_name:
            path_info = env['PATH_INFO']
            if path_info.startswith(script_name): # and it better!
                env['PATH_INFO'] = path_info[len(script_name):]
            else:
                print
                print "All paths are expected to start with the script_name."
                print "  The script_name is %r." % script_name
                print "  The requested path is %r." % self.path
                print
        if self.headers.typeheader is None:
            env['CONTENT_TYPE'] = self.headers.type
        else:
            env['CONTENT_TYPE'] = self.headers.typeheader
        env['CONTENT_LENGTH'] = self.headers.getheader(
            'content-length') or "0"
        for name, value in self.headers.items():
            header_name = 'HTTP_' + name.upper().replace('-', '_')
            env[header_name] = value
        accept = []
        for line in self.headers.getallmatchingheaders('accept'):
            if line[:1] in "\t\n\r ":
                accept.append(line.strip())
            else:
                accept = accept + line[7:].split(',')
        env['HTTP_ACCEPT'] = ','.join(accept)
        co = filter(None, self.headers.getheaders('cookie'))
        if co:
            env['HTTP_COOKIE'] = ', '.join(co)
        env.update(self.required_cgi_environment)
        return env

    def process(self, env):
        hit = get_publisher().process(self.rfile, env)
        response = hit.get_response()
        try:
            self.send_response(*response.get_status())
            response.write(self.wfile, include_status=False,
                           include_body=(self.command!='HEAD'))
        except (IOError, socket.error), err:
            print "Error while sending response ignored: %s" % err

    def handle_one_request(self):
        self.raw_requestline = self.rfile.readline()
        if not self.raw_requestline:
            self.close_connection = 1
            return
        if not self.parse_request():
            return
        return self.process(self.get_cgi_env(self.command))

    def log_request(self, *args, **kwargs):
        pass

    def send_response(self, code, message=None):
        """
        Copied, with regret, from BaseHTTPRequestHandler, except that the line
        that adds the 'Date' header is removed.
        """
        self.log_request(code)
        if message is None:
            if code in self.responses:
                message = self.responses[code][0]
            else:
                message = ''
        if self.request_version != 'HTTP/0.9':
            self.wfile.write("%s %d %s\r\n" %
                             (self.protocol_version, code, message))
        self.send_header('Server', self.version_string())


# netstring utility functions
def ns_read_size(input):
    size = ""
    while 1:
        c = input.read(1)
        if c == ':':
            break
        elif not c:
            if not size:
                return None
            raise IOError, 'short netstring read'
        size = size + c
    return long(size)

def ns_reads(input):
    size = ns_read_size(input)
    if size is None:
        return None
    data = ""
    while size > 0:
        s = input.read(size)
        if not s:
            raise IOError, 'short netstring read'
        data = data + s
        size -= len(s)
    if input.read(1) != ',':
        raise IOError, 'missing netstring terminator'
    return data

def read_env(input):
    headers = ns_reads(input)
    if headers is None:
        return None
    items = headers.split("\0")
    items = items[:-1]
    assert len(items) % 2 == 0, "malformed headers"
    env = {}
    for i in range(0, len(items), 2):
        env[items[i]] = items[i+1]
    return env


class SCGIHandler(object):

    def __init__(self, script_name=''):
        self.script_name = script_name

    def handle_connection(self, conn):
        input = conn.makefile("r")
        output = conn.makefile("w")
        env = read_env(input)
        if env is None:
            # Nothing here.
            return
        if (env.get('SCRIPT_NAME') and
            env.get('SCRIPT_NAME') == env.get('REQUEST_URI') and
            env.get('SCRIPT_NAME').startswith(self.script_name) and
            env.get('PATH_INFO') is None):
            # This looks like it is coming through mod_scgi and
            # needs repair.
            env['PATH_INFO'] = env['SCRIPT_NAME'][len(self.script_name):]
            env['SCRIPT_NAME'] = self.script_name
            assert env['SCRIPT_NAME'] + env['PATH_INFO'] == env['REQUEST_URI']
        hit = get_publisher().process(input, env)
        response = hit.get_response()
        try:
            response.write(output)
            input.close()
            output.close()
            conn.close()
        except IOError, err:
            print "IOError while sending response: %s" % err


def run_web(site):
    logfile = open(site.get_logfile(), 'a')
    sys.stderr = sys.stdout = logfile
    site.start_durus_logging()
    site.get_publisher() # Make sure that the Publisher constructor works.
    clear_publisher()
    scgi_address = site.get_scgi_address()
    http_address = site.get_http_address()
    as_https_address = site.get_as_https_address()
    https_address = site.get_https_address()
    scgi_port = scgi_address and scgi_address[1]
    http_port = http_address and http_address[1]
    as_https_port = as_https_address and as_https_address[1]
    script_name = site.get_script_name()
    def create_connection_handler():
        class ConnectionHandler(object):
            def handle_connection(self, connection):
                port = connection.getsockname()[1]
                if port == scgi_port:
                    return SCGIHandler(
                        script_name=script_name).handle_connection(connection)
                elif port == http_port:
                    self.server_name, self.server_port = http_address
                    client_address = connection.getpeername()
                    HTTPRequestHandler(connection, client_address, self,
                                       SCRIPT_NAME=script_name)
                elif port == as_https_port:
                    self.server_name, self.server_port = https_address
                    client_address = connection.getpeername()
                    HTTPRequestHandler(
                        connection, client_address, self, HTTPS='on',
                        SCRIPT_NAME=script_name)
                else:
                    connection.send(str(locals()))
        site.get_create_publisher()()
        return ConnectionHandler()
    server = HubServer(create_connection_handler,
                       max_children=site.get_max_children(),
                       banned=site.get_banned())
    server.listen(scgi_address, http_address, as_https_address)
    site.ensure_uid_gid_not_root()
    def reexec(*args):
        server.do_stop()
        source = __file__
        if source.endswith('.pyc') or source.endswith('.pyo'):
            source = source[:-1]
        argv = [sys.executable, source, site.get_name()]
        print "RESTART:", " ".join(argv)
        os.execv(sys.executable, argv)
    signal(SIGHUP, reexec)
    server.run()

if __name__ == '__main__':
    run_web(Site(sys.argv[1]))





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/setup.py $
$Id: setup.py 29024 2006-11-01 15:57:20Z dbinger $
"""
import os
import sys
from os import curdir
from os.path import join, exists
from distutils.core import setup
from distutils.extension import Extension
from shutil import copytree

try:
    from qpy.setup import qpy_build_py
except ImportError:
    print "Please install Qpy first."
    raise SystemExit

from __init__ import __version__

if 'sdist' in sys.argv:
    import qpy
    import durus
    import sancho
    import re

    # Make sure that version numbers have all been updated.
    PAT = re.compile(r'\b%s\b' % re.escape(__version__))
    assert len(PAT.findall(open("LICENSE.txt").read())) == 14, __version__
    assert PAT.search(open("CHANGES.txt").readline()), __version__
    assert len(PAT.findall(open("README.txt").read())) == 3, __version__
    QPYPAT = re.compile(r'\b%s\b' % re.escape(qpy.__version__))
    assert len(QPYPAT.findall(open("README.txt").read())) == 3, qpy.__version__
    DURUSPAT = re.compile(r'\b%s\b' % re.escape(durus.__version__))
    assert len(DURUSPAT.findall(open("README.txt").read())) == 3, durus.__version__
    SANCHOPAT = re.compile(r'\b%s\b' % re.escape(sancho.__version__))
    assert len(SANCHOPAT.findall(open("README.txt").read())) == 1, sancho.__version__    

    # Make sure that copyright statements are current.
    from datetime import datetime
    year = datetime.now().year
    copyright = "Copyright (c) Corporation for National Research Initiatives %s" % year
    assert open("README.txt").read().count(copyright) == 1

if __name__ == '__main__':
    if len(sys.argv) == 3 and sys.argv[1] == 'install_demo_sites':
        sites_directory = sys.argv[2]
        if exists(sites_directory):
            print sites_directory, 'already exists.'
            print "To install the demo sites, copy them "
            print "from the 'demo' directory into %r" % sites_directory
        else:
            copytree('demo', sites_directory)
            f = open(join(sites_directory, "__init__.py"), "w")
            f.close()
    else:
        if os.name == 'posix':
            Passfd = Extension(name="qp.hub.passfd", sources=['hub/passfd.c'])
            ext_modules = [Passfd]
        else:
            ext_modules = []
        setup(name="qp",
              version=__version__,
              author="CNRI",
              author_email="webmaster@mems-exchange.org",
              url="http://www.mems-exchange.org/software/qp/",
              package_dir=dict(qp=curdir),
              scripts=['bin/qp', 'bin/qpcensus.py'],
              license="See LICENSE.txt",
              packages=['qp',
                        'qp.http',
                        'qp.hub',
                        'qp.pub',
                        'qp.fill',
                        'qp.lib',
                        'qp.mail',
                        'qp.sites',
                        ],
              ext_modules = ext_modules,
              description="A web-application framework.",
              cmdclass= {'build_py': qpy_build_py},
              long_description=(
            "QP is a package for defining and running multiple web "
            "applications based on Durus for persistence, standard persistent "
            "Session and User classes, easy interactive database sessions, "
            "qpy for assembling html, and Quixote2-style forms and path "
            "traversal.  ")
              )



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/fill/__init__.py $
$Id: __init__.py 26983 2005-06-29 18:57:59Z dbinger $
"""
from qpy.compile import compile_qpy_files
compile_qpy_files(__path__[0])



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/http/response.py $
$Id: response.py 27682 2005-11-08 16:43:55Z dbinger $
"""
from rfc822 import formatdate
import struct
import time
import zlib

status_reasons = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Moved Temporarily',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Time-out',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Large',
    415: 'Unsupported Media Type',
    416: 'Requested range not satisfiable',
    417: 'Expectation Failed',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Time-out',
    505: 'HTTP Version not supported',
    507: 'Insufficient Storage',
}

_GZIP_HEADER = ("\037\213" # magic
                "\010" # compression method
                "\000" # flags
                "\000\000\000\000" # time, who cares?
                "\002"
                "\377")

_GZIP_EXCLUDE = set(["application/pdf",
                     "application/zip",
                     "audio/mpeg",
                     "image/gif",
                     "image/jpeg",
                     "image/png",
                     "video/mpeg",
                     "video/quicktime",
                     "video/x-msvideo",
                     ])


class HTTPResponse:
    """
    An object representation of an HTTP response.

    The Response type encapsulates all possible responses to HTTP
    requests.

    Instance attributes:
      content_type : (mime_type:str, charset:str)
      status : (status_code:int, reason_phrase:str)
      headers : { string : string }
        most of the headers included with the response; every header set
        by 'set_header()' goes here.  Does not include "Status" or
        "Set-Cookie" headers (unless someone uses set_header() to set
        them, but that would be foolish).
      body : str | Stream
        the response body, None by default.  Note that if the body is not a
        stream then it is already encoded using 'charset'.
      buffered : bool
        if false, response data will be flushed as soon as it is
        written (the default is true).  This is most useful for
        responses that use the Stream() protocol.  Note that whether the
        client actually receives the partial response data is highly
        dependent on the web server
      cookies : { name:string : { attrname : value } }
        collection of cookies to set in this response; it is expected
        that the user-agent will remember the cookies and send them on
        future requests.  The cookie value is stored as the "value"
        attribute.  The other attributes are as specified by RFC 2109.
      cache : int | None
        the number of seconds the response may be cached.  The default is 0,
        meaning don't cache at all.  This variable is used to set the HTTP
        expires header.  If set to None then the expires header will not be
        added.
      compress : bool
        should the body of the response be compressed?
    """

    def __init__(self):
        self.set_content_type('text/html', 'utf-8')
        self.set_status(200)
        self.set_compress(False)
        self.set_expires(0)
        self.set_buffered(True)
        self.headers = {}
        self.cookies = {}
        self.body = None

    def set_content_type(self, mime_type, charset):
        """(mime_type:basicstr, charset:basicstr)
        """
        self.content_type = (mime_type, charset)

    def get_mime_type(self):
        return self.content_type[0]

    def get_charset(self):
        return self.content_type[1]

    def set_compress(self, value):
        self.compress = value

    def get_compress(self):
        return self.compress

    def set_buffered(self, value):
        self.buffered = value

    def get_buffered(self):
        return self.buffered

    def set_status(self, status_code, reason=None):
        """(status_code : int, reason : string = None)

        Sets the HTTP status code of the response.  'status_code' must be an
        integer in the range 100 .. 599.  'reason' must be a string; if
        not supplied, the default reason phrase for 'status_code' will be
        used.  If 'status_code' is a non-standard status code, the generic
        reason phrase for its group of status codes will be used; eg.
        if status == 493, the reason for status 400 will be used.
        """
        if not isinstance(status_code, int):
            raise TypeError, "status_code must be an integer"
        if not (100 <= status_code <= 599):
            raise ValueError, "status_code must be between 100 and 599"
        self.status = (status_code,
                       (reason or
                        status_reasons.get(status_code) or
                        status_reasons.get(status_code - (status_code % 100))))

    def get_status(self):
        return self.status

    def get_status_code(self):
        return self.status[0]

    def set_header(self, name, value):
        """(name : string, value : string)"""
        self.headers[name] = value

    def get_header(self, name, default=None):
        """(name : string, default=None) -> value : string

        Gets an HTTP return header "name".  If none exists then 'default' is
        returned.
        """
        lower_name = name.lower()
        for header in self.headers:
            if header.lower() == lower_name:
                return self.headers[header]
        return default

    def set_expires(self, seconds=0, minutes=0, hours=0, days=0):
        if seconds is None:
            self.cache = None # don't generate 'Expires' header
        else:
            self.cache = seconds + 60*(minutes + 60*(hours + 24*days))

    def _compress_body(self, body):
        """(body: str) -> str
        Try to compress the body using gzip and return either the original
        body or the compressed body.  If the compressed body is returned, the
        content-encoding header is set to 'gzip'.
        """
        if not self.get_compress() or self.get_mime_type() in _GZIP_EXCLUDE:
            return body
        n = len(body)
        co = zlib.compressobj(6, zlib.DEFLATED, -zlib.MAX_WBITS,
                              zlib.DEF_MEM_LEVEL, 0)
        chunks = [_GZIP_HEADER,
                  co.compress(body),
                  co.flush(),
                  struct.pack("<ll", zlib.crc32(body), n)]
        compressed_body = "".join(chunks)
        ratio = float(n) / len(compressed_body)
        #print "gzip original size %d, ratio %.1f" % (n, ratio)
        if ratio > 1.0:
            self.set_header("Content-Encoding", "gzip")
            return compressed_body
        else:
            return body

    def _encode_chunk(self, chunk):
        """(chunk:str|unicode)
        Returns the chunk encoded using the charset of the response.
        If the chunk is a str, it is assumed to be encoded already
        using the charset of the response.
        """
        if isinstance(chunk, unicode):
            return chunk.encode(self.get_charset())
        else:
            return chunk

    def set_body(self, body):
        """(body : Stream|basestring)

        Sets the response body equal to the argument 'body'.
        If the argument is a str, or a Stream that generates str instances,
        it is the caller's responsibility to make sure that these str instances
        use the charset of the response.
        """
        if isinstance(body, Stream):
            self.body = body
        else:
            encoded = self._encode_chunk(body)
            self.body = self._compress_body(encoded)

    def expire_cookie(self, name, **attrs):
        """
        Cause an HTTP cookie to be removed from the browser

        The response will include an HTTP header that will remove the cookie
        corresponding to "name" on the client, if one exists.  This is
        accomplished by sending a new cookie with an expiration date
        that has already passed.  Note that some clients require a path
        to be specified - this path must exactly match the path given
        when creating the cookie.  The path can be specified as a keyword
        argument.
        """
        params = {'max_age': 0, 'expires': 'Thu, 01-Jan-1970 00:00:00 GMT'}
        params.update(attrs)
        self.set_cookie(name, "deleted", **params)

    def set_cookie(self, name, value, **attrs):
        """(name : string, value : string, **attrs)

        Set an HTTP cookie on the browser.

        The response will include an HTTP header that sets a cookie on
        cookie-enabled browsers with a key "name" and value "value".
        Cookie attributes such as "expires" and "domains" may be
        supplied as keyword arguments; see RFC 2109 for a full list.
        (For the "secure" attribute, use any true value.)

        This overrides any previous value for this cookie.  Any
        previously-set attributes for the cookie are preserved, unless
        they are explicitly overridden with keyword arguments to this
        call.
        """
        cookies = self.cookies
        if cookies.has_key(name):
            cookie = cookies[name]
        else:
            cookie = cookies[name] = {}
        cookie.update(attrs)
        cookie['value'] = value

    def get_content_length(self):
        if self.body is None:
            return None
        if isinstance(self.body, Stream):
            return self.body.get_length()
        return len(self.body)

    def _gen_cookie_headers(self):
        """() -> [(str, str)]
        """
        for name, attrs in self.cookies.iteritems():
            value = str(attrs['value'])
            if '"' in value:
                value = value.replace('"', '\\"')
            chunks = ['%s="%s"' % (name, value)]
            for name, val in attrs.items():
                name = name.lower()
                if val is None:
                    continue
                if name in ('expires', 'domain', 'path', 'max_age', 'comment'):
                    name = name.replace('_', '-')
                    chunks.append('%s=%s' % (name, val))
                elif name == 'secure' and val:
                    chunks.append("secure")
            yield ("Set-Cookie", '; '.join(chunks))


    def generate_headers(self):
        """() -> [(name:string, value:string)]

        Generate a list of headers to be returned as part of the response.
        """
        for name, value in self.headers.iteritems():
            yield (name.title(), value)

        for name, value in self._gen_cookie_headers():
            yield name, value

        # Date header
        now = time.time()
        if "date" not in self.headers:
            yield ("Date", formatdate(now))

        # Cache directives
        if self.cache is None:
            pass # don't mess with the expires header
        elif "expires" not in self.headers:
            if self.cache > 0:
                expire_date = formatdate(now + self.cache)
            else:
                expire_date = "-1" # allowed by HTTP spec and may work better
                                   # with some clients
            yield ("Expires", expire_date)

        # Content-type
        if "content-type" not in self.headers:
            mime_type, charset = self.content_type
            value = mime_type
            if charset:
                value += '; charset=%s' % charset
            yield ('Content-Type', value)

        # Content-Length
        if "content-length" not in self.headers:
            length = self.get_content_length()
            if length is not None:
                yield ('Content-Length', str(length))


    def generate_body_chunks(self):
        """
        Return a sequence of body chunks, encoded using 'charset'.
        Note that the chunks in the iteration of a Stream, if they
        are str instances, are assumed to be encoded already.
        """
        if self.body is None:
            pass
        elif isinstance(self.body, Stream):
            for chunk in self.body:
                yield self._encode_chunk(chunk)
        else:
            yield self.body # already encoded by set_body().

    def write(self, output, include_status=True, include_body=True):
        """(output:file, include_status:bool=True, include_body:bool=True)

        Write the HTTP response headers and body to 'output'.  This is not
        a complete HTTP response, as it doesn't start with a response
        status line as specified by RFC 2616.  By default, it does start
        with a "Status" header as described by the CGI spec.  It is expected
        that this response is parsed by the web server and turned into a
        complete HTTP response.  If include_body is False, only the headers
        are written to 'output'.  This is used to support HTTP HEAD requests.
        """
        flush_output = not self.get_buffered() and hasattr(output, 'flush')
        if include_status:
            # "Status" header must come first.
            output.write("Status: %03d %s\r\n" % self.status)
        for name, value in self.generate_headers():
            output.write("%s: %s\r\n" % (name, value))
        output.write("\r\n")
        if flush_output:
            output.flush()
        if not include_body:
            return
        for chunk in self.generate_body_chunks():
            output.write(chunk)
            if flush_output:
                output.flush()
        if flush_output:
            output.flush()

class Stream:
    """
    A wrapper around response data that can be streamed.  The 'iterable'
    argument must support the iteration protocol.
    Beware that exceptions raised while writing the stream will not be
    handled gracefully.

    Instance attributes:
      iterable : any
        an object that supports the iteration protocol.  The items produced
        by the stream must be be unicode or str instances.
        If they are str instances, the encoding is assumed to be that of
        the response.
      length: int | None
        the number of bytes that will be produced by the stream, None
        if it is not known.  Used to set the Content-Length header.
    """
    def __init__(self, iterable, length=None):
        self.iterable = iterable
        self.length = length

    def __iter__(self):
        return iter(self.iterable)

    def get_length(self):
        return self.length





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/common.py $
$Id: common.py 28673 2006-08-30 00:58:21Z dbinger $

Code for maintaining the global Publisher, and convenience functions
for operations that go through the publisher.
"""
from threading import _get_ident

_publisher = {}  #  { int : Publisher }

def set_publisher(value):
    global _publisher
    assert _get_ident() not in _publisher
    _publisher[_get_ident()] = value

def clear_publisher():
    # Used in test scripts only.
    global _publisher
    if _get_ident() in _publisher:
        del _publisher[_get_ident()]

def get_publisher():
    return _publisher.get(_get_ident())

# All of the remaining functions here work through get_publisher().

def get_hit():
    return get_publisher().get_hit()

def get_request():
    return get_hit().get_request()

def get_response():
    return get_hit().get_response()

def get_site():
    return get_publisher().get_site()

def get_path(n=0):
    return get_request().get_path(n=n)

def redirect(location, permanent=False):
    get_publisher().redirect(location, permanent=permanent)

def not_found(body=None):
    get_publisher().not_found(body=body)

def header(title, *args, **kwargs):
    return get_publisher().header(title, *args, **kwargs)

def footer(*args, **kwargs):
    return get_publisher().footer(*args, **kwargs)

def page(title, *content, **kwargs):
    return get_publisher().page(title, *content, **kwargs)

def respond(title, *content, **kwargs):
    return get_publisher().respond(title, *content, **kwargs)

def site_now():
    return get_publisher().get_time_zone().now()

# The remaining are for sites using DurusPublisher.

def get_session():
    return get_hit().get_session()

def get_user():
    return get_session().get_effective_user()

def ensure_signed_in(**kwargs):
    return get_publisher().ensure_signed_in(**kwargs)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/tz.py $
$Id: tz.py 28672 2006-08-30 00:53:48Z dbinger $

Adapted from http://www.python.org/doc/2.3.5/lib/tzinfo-examples.txt
"""
from datetime import tzinfo, timedelta, datetime
from types import FunctionType

class ClassMethodsClass (type):
    """
    A class that uses this class as the value of its __metaclass__ attribute
    gets all of its methods (except those with names that start with '__')
    converted into class methods.  This is a way to define
    classes that can act as singleton instances.

    The str method of classes with this metaclass returns the name of
    the class.
    """

    def __init__ (self, class_name, bases, namespace):
        for name, obj in namespace.items():
            if isinstance(obj, FunctionType) and not name.startswith('__'):
                setattr(self, name, classmethod(obj))

    def __str__ (self):
        return self.__name__


class _tzinfoclass (ClassMethodsClass, tzinfo):
    """
    Classes with this class are *instances* of tzinfo. (!)
    All of the non __* methods are automatically converted
    to classmethods.
    """

class TimeZone (object):

    __metaclass__ = _tzinfoclass

    def now(klass):
        """
        The current time, with this timezone.
        """
        return datetime.now(klass)


class UTCOffset (TimeZone):

    _utc_offset = timedelta(0)
    _name = "UTC?"
    _dst = timedelta(0)

    def utcoffset(klass, dt):
        return klass._utc_offset

    def tzname(klass, dt):
        return klass._name

    def dst(klass, dt):
        return klass._dst

class UTC (UTCOffset):
    _name = "UTC"


_ZERO_TIME = timedelta(hours=0)
_HOUR = timedelta(hours=1)

class _US (TimeZone):

    # In the US, DST starts at 2am (standard time) on the first Sunday in April.
    _DSTSTART = datetime(1, 4, 1, 2)
    # and ends at 2am (DST time; 1am standard time) on the last Sunday of Oct.
    # which is the first Sunday on or after Oct 25.
    _DSTEND = datetime(1, 10, 25, 1)

    def tzname(klass, dt):
        if klass.dst(dt):
            return klass.dstname
        else:
            return klass.stdname

    def utcoffset(klass, dt):
        return klass.stdoffset + klass.dst(dt)

    def dst(klass, dt):
        if dt is None or dt.tzinfo is None:
            # An exception may be sensible here, in one or both cases.
            # It depends on how you want to treat them.  The default
            # fromutc() implementation (called by the default astimezone()
            # implementation) passes a datetime with dt.tzinfo is klass.
            return _ZERO_TIME
        assert dt.tzinfo is klass

        def first_sunday_on_or_after(dt):
            days_to_go = 6 - dt.weekday()
            if days_to_go:
                dt += timedelta(days_to_go)
            return dt

        # Find first Sunday in April & the last in October.
        start = first_sunday_on_or_after(klass._DSTSTART.replace(year=dt.year))
        end = first_sunday_on_or_after(klass._DSTEND.replace(year=dt.year))

        # Can't compare naive to aware objects, so strip the timezone from
        # dt first.
        if start <= dt.replace(tzinfo=None) < end:
            return _HOUR
        else:
            return _ZERO_TIME

class Eastern (_US):
    stdoffset = timedelta(hours=-5)
    stdname = "EST"
    dstname = "EDT"

class Central (_US):
    stdoffset = timedelta(hours=-6)
    stdname = "CST"
    dstname = "CDT"

class Mountain (_US):
    stdoffset = timedelta(hours=-7)
    stdname = "MST"
    dstname = "MDT"

class Pacific (_US):
    stdoffset = timedelta(hours=-8)
    stdname = "PST"
    dstname = "PDT"

class UTC_plus_1 (UTCOffset):
    _name = "UTC+1"
    _utc_offset = timedelta(hours=1)

class UTC_minus_1 (UTCOffset):
    _name = "UTC-1"
    _utc_offset = timedelta(hours=-1)

class UTC_plus_2 (UTCOffset):
    _name = "UTC+2"
    _utc_offset = timedelta(hours=2)

class UTC_minus_2 (UTCOffset):
    _name = "UTC-2"
    _utc_offset = timedelta(hours=-2)

class UTC_plus_3 (UTCOffset):
    _name = "UTC+3"
    _utc_offset = timedelta(hours=3)

class UTC_minus_3 (UTCOffset):
    _name = "UTC-3"
    _utc_offset = timedelta(hours=-3)

class UTC_plus_4 (UTCOffset):
    _name = "UTC+4"
    _utc_offset = timedelta(hours=4)

class UTC_minus_4 (UTCOffset):
    _name = "UTC-4"
    _utc_offset = timedelta(hours=-4)

class UTC_plus_5 (UTCOffset):
    _name = "UTC+5"
    _utc_offset = timedelta(hours=5)

class UTC_minus_5 (UTCOffset):
    _name = "UTC-5"
    _utc_offset = timedelta(hours=-5)

class UTC_plus_6 (UTCOffset):
    _name = "UTC+6"
    _utc_offset = timedelta(hours=6)

class UTC_minus_6 (UTCOffset):
    _name = "UTC-6"
    _utc_offset = timedelta(hours=-6)

class UTC_plus_7 (UTCOffset):
    _name = "UTC+7"
    _utc_offset = timedelta(hours=7)

class UTC_minus_7 (UTCOffset):
    _name = "UTC-7"
    _utc_offset = timedelta(hours=-7)

class UTC_plus_8 (UTCOffset):
    _name = "UTC+8"
    _utc_offset = timedelta(hours=8)

class UTC_minus_8 (UTCOffset):
    _name = "UTC-8"
    _utc_offset = timedelta(hours=-8)

class UTC_plus_9 (UTCOffset):
    _name = "UTC+9"
    _utc_offset = timedelta(hours=9)

class UTC_minus_9 (UTCOffset):
    _name = "UTC-9"
    _utc_offset = timedelta(hours=-9)

class UTC_plus_10 (UTCOffset):
    _name = "UTC+10"
    _utc_offset = timedelta(hours=10)

class UTC_minus_10 (UTCOffset):
    _name = "UTC-10"
    _utc_offset = timedelta(hours=-10)

class UTC_plus_11 (UTCOffset):
    _name = "UTC+11"
    _utc_offset = timedelta(hours=11)

class UTC_minus_11 (UTCOffset):
    _name = "UTC-11"
    _utc_offset = timedelta(hours=-11)

class UTC_plus_12 (UTCOffset):
    _name = "UTC+12"
    _utc_offset = timedelta(hours=12)

class UTC_minus_12 (UTCOffset):
    _name = "UTC-12"
    _utc_offset = timedelta(hours=-12)

def now(tzinfo=UTC):
    return datetime.now(tzinfo=tzinfo)



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/http/test/utest_response.py $
$Id: utest_response.py 27727 2005-12-13 15:19:03Z dbinger $
"""
from cStringIO import StringIO
from qp.http.response import HTTPResponse, Stream
from sancho.utest import UTest

class Test(UTest):

    def a(self):
        response = HTTPResponse()
        response.set_content_type('text/html', 'utf8')
        response.set_status(200)
        response.set_status(200, 'Excellent')
        assert response.get_status_code() == 200
        assert response.get_status() == (200, 'Excellent')
        response.set_status(479)
        try:
            response.set_status(600)
            assert 0
        except ValueError:
            pass
        try:
            response.set_status('200')
            assert 0
        except TypeError:
            pass
        response.set_expires(None)
        response.set_expires(seconds=200)
        response.set_header('foo', 1)
        assert response.get_header('foo') == 1
        response.set_cookie('a', 23)
        response.expire_cookie('a')
        response.set_body('ok')
        response.set_compress(False)
        response.set_body('ok' * 1000)

    def b(self):
        response = HTTPResponse()
        assert response.get_mime_type() == 'text/html'
        response.set_content_type('text/plain', None)
        assert response.get_mime_type() == 'text/plain'
        assert (200, 'OK') == response.status
        assert response.get_content_length() == None
        response.set_body('ab')
        assert response.get_content_length() == 2
        response.set_body(Stream(StringIO('a'), 1))
        assert response.get_content_length() == 1
        response.set_cookie('cookie', 1)
        headers = response.generate_headers()
        headers_dict = dict(headers)
        del headers_dict['Date']
        print headers_dict
        assert headers_dict == dict([
            ('Content-Length', '1'),
            ('Content-Type', 'text/plain'),
            ('Expires', '-1'),
            ('Set-Cookie', 'cookie="1"')])
        response.set_expires(None)
        headers = response.generate_headers()
        response.set_expires(days=1)
        headers = response.generate_headers()
        response.set_cookie('a', 'x"', secure=1, expires=1,
                            b=None, a='all"ok"')
        response.set_header('bogus', 'yes')
        headers = response.generate_headers()

    def d(self):
        response = HTTPResponse()
        response.set_cookie('a', 'x')
        headers = response.generate_headers()
        headers_dict = dict(headers)
        del headers_dict['Date']
        print headers_dict
        assert headers_dict == dict([
            ('Content-Type', 'text/html; charset=utf-8'),
            ('Expires', '-1'),
            ('Set-Cookie', 'a="x"')])

    def check_write(self):
        stream = Stream(StringIO('a'), 1)
        assert ''.join(stream) == 'a'
        response = HTTPResponse()
        response.buffered=False
        out = StringIO()
        response.write(out, include_status=False)
        response.write(out, include_status=True)
        response.write(out, include_body=False)
        response.set_body('a')
        response.write(out)
        response.set_body(stream)
        response.write(out)
        response.set_body(Stream(StringIO('a'*1000), 1000))
        response.write(out)
        response.set_body(u'0' * 1000)
        response.write(out)

    def c(self):
        response = HTTPResponse()
        response.set_cookie('a', '"ok', path="/", secure=True)
        response.set_cookie('b', None)
        response.set_header('ok', 'this')
        list(response.generate_headers())
        assert response.cache == 0
        response.set_expires(1)
        list(response.generate_headers())
        response.cache = None
        list(response.generate_headers())

if __name__ == '__main__':
    Test()
